#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides widgets that display the database in a tree format"""

# todo: tango-centric!!

__all__ = ["TaurusDbTreeWidget"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.core.taurusbasetypes import TaurusElementType
from taurus.core.taurusauthority import TaurusAuthority
from taurus.qt.qtcore.model import *
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.icon import getElementTypeIcon, getElementTypeIconName
from taurustree import TaurusBaseTreeWidget


class TaurusDbTreeWidget(TaurusBaseTreeWidget):
    """A class:`taurus.qt.qtgui.tree.TaurusBaseTreeWidget` that connects to a
    :class:`taurus.core.taurusauthority.TaurusAuthority` model. It can show the list of database
    elements in four different perspectives:

    - device : a three level hierarchy of devices (domain/family/name)
    - server : a server based perspective
    - class : a class based perspective

    Filters can be inserted into this widget to restrict the tree nodes that are
    seen.
    """

    KnownPerspectives = {
        TaurusElementType.Device: {
            "label": "By device",
            "icon": getElementTypeIconName(TaurusElementType.Device),
            "tooltip": "View by device tree",
            "model": [TaurusDbDeviceProxyModel, TaurusDbDeviceModel, ],
        },
        'PlainDevice': {
            "label": "By plain device",
            "icon": getElementTypeIconName(TaurusElementType.Device),
            "tooltip": "View by plain device tree (it may take a long time if there are problems with the exported devices)",
            "model": [TaurusDbDeviceProxyModel, TaurusDbPlainDeviceModel, ],
        },

        TaurusElementType.Server: {
            "label": "By server",
            "icon": getElementTypeIconName(TaurusElementType.Server),
            "tooltip": "View by server tree",
            "model": [TaurusDbServerProxyModel, TaurusDbServerModel, ],
        },
        TaurusElementType.DeviceClass: {
            "label": "By class",
            "icon": getElementTypeIconName(TaurusElementType.DeviceClass),
            "tooltip": "View by class tree",
            "model": [TaurusDbDeviceClassProxyModel, TaurusDbDeviceClassModel, ],
        },
    }

    DftPerspective = TaurusElementType.Device

    def getModelClass(self):
        return TaurusAuthority

    def sizeHint(self):
        return Qt.QSize(1024, 512)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.tree'
        ret['group'] = 'Taurus Views'
        ret['icon'] = "designer:listview.png"
        return ret


class _TaurusTreePanel(Qt.QWidget, TaurusBaseWidget):
    """A demonstration panel to show how :class:`taurus.qt.qtcore.TaurusDbBaseModel`
    models can interact with several model view widgets like QTreeView,
    QTableView, QListView and QComboBox"""

    def __init__(self, parent=None, designMode=False):
        """doc please!"""
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QWidget, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        self.init(designMode)

    def init(self, designMode):
        l = Qt.QGridLayout()
        l.setContentsMargins(0, 0, 0, 0)
        self.setLayout(l)

#        tb = self._toolbar = Qt.QToolBar("Taurus tree panel toolbar")
#        tb.setFloatable(False)
#        refreshAction = self._refreshAction = tb.addAction(Qt.QIcon.fromTheme("view-refresh"),"Refresh", self.refresh)

#        l.addWidget(tb, 0, 0)

        main_panel = Qt.QTabWidget()
        self._device_tree_view = TaurusDbTreeWidget(
            perspective=TaurusElementType.Device)
        self._device_table_view = Qt.QTableView()
        self._device_table_view.setModel(TaurusDbBaseModel())
        self._device_list_view = Qt.QListView()
        self._device_list_view.setModel(TaurusDbSimpleDeviceModel())
        self._server_tree_view = TaurusDbTreeWidget(
            perspective=TaurusElementType.Server)
        self._class_tree_view = TaurusDbTreeWidget(
            perspective=TaurusElementType.DeviceClass)

        self._device_combo_view = Qt.QWidget()
        combo_form = Qt.QFormLayout()
        self._device_combo_view.setLayout(combo_form)

        self._combo_dev_tree_widget = TaurusDbTreeWidget(
            perspective=TaurusElementType.Device)
        qmodel = self._combo_dev_tree_widget.getQModel()
        qmodel.setSelectables([TaurusElementType.Member])
        device_combo = Qt.QComboBox()
        device_combo.setModel(qmodel)
        device_combo.setMaxVisibleItems(20)
        device_combo.setView(self._combo_dev_tree_widget.treeView())
        combo_form.addRow(
            "Device selector (by device hierarchy):", device_combo)

        self._combo_attr_tree_widget = TaurusDbTreeWidget(
            perspective=TaurusElementType.Device)
        qmodel = self._combo_attr_tree_widget.getQModel()
        device_combo = Qt.QComboBox()
        device_combo.setModel(qmodel)
        device_combo.setMaxVisibleItems(20)
        device_combo.setView(self._combo_attr_tree_widget.treeView())
        combo_form.addRow(
            "Attribute selector (by device hierarchy):", device_combo)

        self._combo_dev_table_view = Qt.QTableView()
        self._combo_dev_table_view.setModel(TaurusDbBaseModel())
        qmodel = self._combo_dev_table_view.model()
        qmodel.setSelectables([TaurusElementType.Device])
        device_combo = Qt.QComboBox()
        device_combo.setModel(qmodel)
        device_combo.setMaxVisibleItems(20)
        device_combo.setView(self._combo_dev_table_view)
        combo_form.addRow("Device selector (by plain device):", device_combo)

        main_panel.addTab(self._device_tree_view, "Device (Tree View)")
        main_panel.addTab(self._device_table_view, "Device (Table View)")
        main_panel.addTab(self._device_list_view, "Device (List View)")
        main_panel.addTab(self._server_tree_view, "Server (Tree View)")
        main_panel.addTab(self._class_tree_view, "Class (Tree View)")
        main_panel.addTab(self._device_combo_view, "ComboBox Views")

        l.addWidget(main_panel, 1, 0)

        self._main_panel = main_panel

    def deviceTreeWidget(self):
        return self._device_tree_view

    def deviceTableWidget(self):
        return self._device_table_view

    def deviceListWidget(self):
        return self._device_list_view

    def serverTreeWidget(self):
        return self._server_tree_view

    def classTreeWidget(self):
        return self._class_tree_view

    def sizeHint(self):
        return Qt.QSize(1024, 512)

    def _updateTreeModels(self):
        db_name, db = self.getModel(), self.getModelObj()

        self._device_tree_view.setModel(db_name)

        model = self._device_table_view.model()
        if model is not None:
            model.setDataSource(db)

        model = self._device_list_view.model()
        if model is not None:
            model.setDataSource(db)

        self._server_tree_view.setModel(db_name)
        self._class_tree_view.setModel(db_name)
        self._combo_dev_tree_widget.setModel(db_name)
        self._combo_attr_tree_widget.setModel(db_name)

        model = self._combo_dev_table_view.model()
        if model is not None:
            model.setDataSource(db)

    def refresh(self):
        db = self.getModelObj()
        if db is None:
            return
        db.refreshCache()
        self._device_tree_view.refresh()
        self._device_table_view.model().refresh()
        self._device_list_view.model().refresh()
        self._server_tree_view.refresh()
        self._class_tree_view.refresh()

    def goIntoTree(self):
        index = self._device_tree_view.currentIndex()
        if index is None:
            return
        #index_parent = index.parent()
        # if index_parent is None:
        #    return
        self._device_tree_view.setRootIndex(index)

    def goUpTree(self):
        index = self._device_tree_view.rootIndex()
        if index is None:
            return
        index_parent = index.parent()
        if index_parent is None:
            return
        self._device_tree_view.setRootIndex(index_parent)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getModelClass(self):
        return TaurusAuthority

    @Qt.pyqtSlot('QString')
    def setModel(self, model):
        """Sets/unsets the model name for this component

        :param model: (str) the new model name"""
        super(_TaurusTreePanel, self).setModel(model)
        self._updateTreeModels()

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    #: This property holds the unique URI string representing the model name
    #: with which this widget will get its data from. The convention used for
    #: the string can be found :ref:`here <model-concept>`.
    #:
    #: In case the property :attr:`useParentModel` is set to True, the model
    #: text must start with a '/' followed by the attribute name.
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`TaurusBaseWidget.getModel`
    #:     * :meth:`TaurusBaseWidget.setModel`
    #:     * :meth:`TaurusBaseWidget.resetModel`
    #:
    #: .. seealso:: :ref:`model-concept`
    model = Qt.pyqtProperty("QString", TaurusBaseWidget.getModel,
                            TaurusBaseWidget.setModel, TaurusBaseWidget.resetModel)


def main_TaurusTreePanel(host):
    w = _TaurusTreePanel()
    w.setWindowIcon(getElementTypeIcon(TaurusElementType.Device))
    w.setWindowTitle("A Taurus Tree Example")
    w.setModel(host)
    w.show()
    return w


def main_TaurusDbTreeWidget(host, perspective=TaurusElementType.Device):
    w = TaurusDbTreeWidget(perspective=perspective)
    w.setWindowIcon(getElementTypeIcon(perspective))
    w.setWindowTitle("A Taurus Tree Example")
    w.setModel(host)
    w.show()
    return w


def demo():
    """DB panels"""
    import taurus
    db = taurus.Authority()
    host = db.getNormalName()
    w = main_TaurusTreePanel(host)
    # w = main_TaurusDbTreeWidget(host, TaurusElementType.Device)

    return w


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Application(app_name="DB model demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")
    w = demo()
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__all__ = ['EvaluationDevice']

from taurus import Factory
from taurus.core.taurusdevice import TaurusDevice
from taurus.core.util.safeeval import SafeEvaluator
from taurus.core.taurusbasetypes import TaurusDevState, TaurusAttrValue


class EvaluationDevice(TaurusDevice, SafeEvaluator):
    """The evaluator object. It is a :class:`TaurusDevice` and is used as the
    parent of :class:`EvaluationAttribute` objects for which it performs the
    mathematical evaluation.

    .. seealso:: :mod:`taurus.core.evaluation`

    .. warning:: In most cases this class should not be instantiated directly.
                 Instead it should be done via the
                 :meth:`EvaluationFactory.getDevice`
    """
    _symbols = []
    # helper class property that stores a reference to the corresponding
    # factory
    _factory = None
    _scheme = 'eval'

    def __init__(self, name, **kw):
        """Object initialization."""
        self.call__init__(TaurusDevice, name, **kw)
        safedict = {}
        for s in self._symbols:
            if hasattr(self, s):
                safedict[s] = getattr(self, s)
        SafeEvaluator.__init__(self, safedict=safedict)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusDevice necessary overwrite
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    def _createHWObject(self):
        return 'Evaluation'

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    def getAttribute(self, attrname):
        """Returns the attribute object given its name"""
        full_attrname = "%s;%s" % (self.getFullName(), attrname)
        return self.factory().getAttribute(full_attrname)

    def decode(self, event_value):
        # TODO: Is this method ever called? (or maybe just garbage from 3.x?)
        if isinstance(event_value, int):  # TaurusSWDevState
            new_sw_state = event_value
        else:
            self.info("Unexpected value to decode: %s" % str(event_value))
            new_sw_state = TaurusDevState.NotReady
        value = TaurusAttrValue()
        value.rvalue = new_sw_state
        return value



# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Taurus
##
## http://taurus-scada.org
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import sys

import taurus
from taurus import tauruscustomsettings
from taurus.external import unittest

_QtAPIs = ["PyQt4", "PySide", "PyQt5"]


def _import(name):
    __import__(name)
    return sys.modules[name]


class QtTestCase(unittest.TestCase):

    QtAPI = None

    def setUp(self):
        taurus.setLogLevel(taurus.Critical)

        for qt in _QtAPIs:
            if qt in sys.modules:
                self.QtAPI = qt
                break
        else:
            self.QtAPI = "PyQt4"

        self.opt_mods = ("QtDesigner", "QtNetwork", "Qt", "QtSvg",
                         "QtUiTools", "QtWebKit", "Qwt5", "uic")

        # store a "snapshot" of the currently loaded modules
        self._orig_mods = set(sys.modules.keys())

        # auto initialize Qt by taurus using forcibly the self.QtAPI
        tauruscustomsettings.QT_AUTO_INIT = True
        tauruscustomsettings.QT_AUTO_API = self.QtAPI
        tauruscustomsettings.QT_AUTO_STRICT = True

        # this import initializes Qt in case it is not loaded
        from taurus.external.qt import Qt
        self.__qt = Qt

    def test_qt_base_import(self):
        mods = set(sys.modules.keys())

        other_apis = set(_QtAPIs)
        other_apis.remove(self.QtAPI)

        from taurus.external.qt import getQtName

        self.assertEquals(getQtName(), self.QtAPI)

        for other_api in other_apis:
            self.assertFalse(other_api in mods, other_api + " loaded in " + self.QtAPI + " test")

        self.assertTrue(self.QtAPI in mods, self.QtAPI + " not loaded")
        self.assertTrue(self.QtAPI + ".QtCore" in mods, "QtCore not loaded")

        for opt_mod in self.opt_mods:
            mod = "{0}.{1}".format(self.QtAPI, opt_mod)
            self.assertFalse(mod in mods-self._orig_mods, mod + " is loaded")

    def __test_qt_module(self, qt_mod_name):
        taurus_qt_mod_name = "taurus.external.qt.{0}".format(qt_mod_name)
        orig_qt_mod_name = "{0}.{1}".format(self.QtAPI, qt_mod_name)
        TaurusQtMod = _import(taurus_qt_mod_name)
        OrigQtMod = _import(orig_qt_mod_name)
        taurus_qt_mod_members = [ m for m in dir(TaurusQtMod) if not m.startswith("_") ]
        orig_qt_mod_members = [ m for m in dir(OrigQtMod) if not m.startswith("_") ]

        for orig_member_name in orig_qt_mod_members:
            self.assertTrue(orig_member_name in taurus_qt_mod_members,
                            "Taurus {0} does not contain {1}".format(qt_mod_name, orig_member_name))

    def test_qt_core(self):
        return self.__test_qt_module("QtCore")

    def test_qt_gui(self):
        return self.__test_qt_module("QtGui")

    def test_icons(self):
        '''check that theme icons work'''
        from taurus.external.qt import Qt
        icon = Qt.QIcon.fromTheme("folder-open")
        msg = ('Theme icons not available ' + 
               '(if PyQt<4.6, make sure to build resources first!)')
        self.assertFalse(icon.isNull(), msg)


def  main():
    unittest.main(verbosity=2)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""A Qt MainWindow for the TaurusConsole

This is a tabbed pseudo-terminal of IPython sessions, with a menu bar for
common actions.
"""

__all__ = ["TaurusConsoleWindow"]

__docformat__ = 'restructuredtext'

import functools

from taurus.external.qt import Qt
try:
    from IPython.qt.console.mainwindow import MainWindow
except ImportError:
    from IPython.frontend.qt.console.mainwindow import MainWindow


class TaurusConsoleWindow(MainWindow):

    def __init__(self, *args, **kwargs):
        self._pending_kernel_actions = []
        self.new_kernel_menu = None
        MainWindow.__init__(self, *args, **kwargs)

    def add_tab_with_frontend(self, frontend, name=None):
        init_menu = self.tab_widget.count() == 0
        super(TaurusConsoleWindow, self).add_tab_with_frontend(
            frontend, name=name)
        if init_menu:
            self.init_menu_bar()

    def remove_menu_action(self, menu, action):
        menu.removeAction(action)
        self.removeAction(action)

    def add_new_kernel_action(self, action):
        self.add_menu_action(self.new_kernel_menu, action)

    def init_file_menu(self):
        super(TaurusConsoleWindow, self).init_file_menu()
        file_menu = self.file_menu
        self.remove_menu_action(file_menu, self.new_kernel_tab_act)

        self.new_kernel_menu = new_kernel_menu = Qt.QMenu(
            self.new_kernel_tab_act.text())
        file_menu.insertMenu(self.slave_kernel_tab_act, new_kernel_menu)

        # self.new_kernel_tab_act.setText("IPython")
        #self.add_menu_action(new_kernel_menu, self.new_kernel_tab_act)

        for kernel_action in self._pending_kernel_actions:
            self.add_new_kernel_action(kernel_action)

    def add_new_tango_action(self):
        tango_action = Qt.QAction("Tango", self,
                                  triggered=self.create_tab_with_new_tango_frontend)
        self.add_new_kernel_action(tango_action)

    def create_tab_with_new_frontend(self, name='ipython', label=None):
        widget = self.new_frontend_factory(name=name)
        self.add_tab_with_frontend(widget, name=label)

    def register_kernel_extension(self, extension):
        f = functools.partial(self.create_tab_with_new_frontend,
                              name=extension.Name,
                              label=extension.Label)
        action = Qt.QAction(extension.Label, self, triggered=f)

        if self.new_kernel_menu is None:
            self._pending_kernel_actions.append(action)
        else:
            self.add_new_kernel_action(action)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a pure Qt widget that displays an image"""

__all__ = ["QPixmapWidget"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt


class QPixmapWidget(Qt.QWidget):
    """This widget displays an image (pixmap). By default the pixmap is
    scaled to the widget size and the aspect ratio is kept.
    The default alignment of the pixmap inside the widget space is horizontal
    left, vertical center."""

    DefaultAlignment = Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter
    DefaultAspectRatioMode = Qt.Qt.KeepAspectRatio
    DefaultTransformationMode = Qt.Qt.SmoothTransformation

    def __init__(self, parent=None, designMode=False):
        self._pixmap = Qt.QPixmap()
        self._pixmapDrawn = None
        self._alignment = self.DefaultAlignment
        self._pixmapAspectRatioMode = self.DefaultAspectRatioMode
        self._pixmapTransformationMode = self.DefaultTransformationMode

        Qt.QWidget.__init__(self, parent)

    def _getPixmap(self):
        if self._pixmapDrawn is None:
            self._pixmapDrawn = self.recalculatePixmap()
        return self._pixmapDrawn

    def recalculatePixmap(self):
        origPixmap = self._pixmap
        if origPixmap.isNull():
            return origPixmap
        return origPixmap.scaled(self.size(), self._pixmapAspectRatioMode,
                                 self._pixmapTransformationMode)

    def _setDirty(self):
        self._pixmapDrawn = None

    def paintEvent(self, paintEvent):
        """Overwrite the paintEvent from QWidget to draw the pixmap"""
        pixmap = self._getPixmap()

        w, h = self.width(), self.height()
        painter = Qt.QPainter(self)
        painter.setRenderHint(Qt.QPainter.Antialiasing)
        pw, ph = pixmap.width(), pixmap.height()
        align = self._alignment
        hAlign = align & Qt.Qt.AlignHorizontal_Mask
        vAlign = align & Qt.Qt.AlignVertical_Mask
        x, y = 0, 0
        if hAlign & Qt.Qt.AlignHCenter:
            x = (w - pw) / 2
        elif hAlign & Qt.Qt.AlignRight:
            x = w - pw
        if vAlign & Qt.Qt.AlignVCenter:
            y = (h - ph) / 2
        elif vAlign & Qt.Qt.AlignBottom:
            y = h - ph
        x, y = max(0, x), max(0, y)
        painter.drawPixmap(x, y, pixmap)

    def resizeEvent(self, event):
        self._setDirty()
        return Qt.QWidget.resizeEvent(self, event)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {
            'module': 'taurus.qt.qtgui.display',
            'group': 'Taurus Display',
            'icon': "designer:graphicsview.png",
            'container': False,
        }

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getPixmap(self):
        """Returns the pixmap.Returns None if no pixmap is set.
        :return: the current pixmap
        :rtype: PyQt4.Qt.QPixmap"""
        return self._pixmap

    def setPixmap(self, pixmap):
        """Sets the pixmap for this widget. Setting it to None disables pixmap
        :param pixmap: the new pixmap
        :type  pixmap: PyQt4.Qt.QPixmap"""
        # make sure to make a copy because of bug in PyQt 4.4. This is actually
        # copying the internal bitmap, just the qpixmap, so there is no performance
        # penalty here
        self._pixmap = Qt.QPixmap(pixmap)
        self._setDirty()
        self.update()

    def resetPixmap(self):
        """Resets the pixmap for this widget."""
        self.setPixmap(Qt.QPixmap())

    def getAspectRatioMode(self):
        """Returns the aspect ratio to apply when drawing the pixmap.
        :return: the current aspect ratio
        :rtype: PyQt4.Qt.AspectRatioMode"""
        return self._pixmapAspectRatioMode

    def setAspectRatioMode(self, aspect):
        """Sets the aspect ratio mode to apply when drawing the pixmap.
        :param pixmap: the new aspect ratio mode
        :type  pixmap: PyQt4.Qt.AspectRatioMode"""
        self._pixmapAspectRatioMode = aspect
        self._setDirty()
        self.update()

    def resetAspectRatioMode(self):
        """Resets the aspect ratio mode to KeepAspectRatio"""
        self.setAspectRatioMode(self.DefaultAspectRatioMode)

    def getTransformationMode(self):
        """Returns the transformation mode to apply when drawing the pixmap.
        :return: the current transformation mode
        :rtype: PyQt4.Qt.TransformationMode"""
        return self._pixmapTransformationMode

    def setTransformationMode(self, transformation):
        """Sets the transformation mode to apply when drawing the pixmap.
        :param pixmap: the new transformation mode
        :type  pixmap: PyQt4.Qt.TransformationMode"""
        self._pixmapTransformationMode = transformation
        self._setDirty()
        self.update()

    def resetTransformationMode(self):
        """Resets the transformation mode to SmoothTransformation"""
        self.setTransformationMode(self.DefaultTransformationMode)

    def getAlignment(self):
        """Returns the alignment to apply when drawing the pixmap.
        :return: the current alignment
        :rtype: PyQt4.Qt.Alignment"""
        return self._alignment

    def setAlignment(self, alignment):
        """Sets the alignment to apply when drawing the pixmap.
        :param pixmap: the new alignment
        :type  pixmap: PyQt4.Qt.Alignment"""
        self._alignment = Qt.Qt.Alignment(alignment)
        self.update()

    def resetAlignment(self):
        """Resets the transformation mode to Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter"""
        self.setAlignment(self.DefaultAlignment)

    #: This property holds the widget's pixmap
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QPixmapWidget.getPixmap`
    #:     * :meth:`QPixmapWidget.setPixmap`
    #:     * :meth:`QPixmapWidget.resetLedStatus`
    pixmap = Qt.pyqtProperty("QPixmap", getPixmap, setPixmap,
                             resetPixmap, doc="the widget's pixmap")

    #: This property holds the widget's pixmap aspect ratio mode
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QPixmapWidget.getAspectRatioMode`
    #:     * :meth:`QPixmapWidget.setAspectRatioMode`
    #:     * :meth:`QPixmapWidget.resetAspectRatioMode`
    aspectRatioMode = Qt.pyqtProperty("Qt::AspectRatioMode", getAspectRatioMode,
                                      setAspectRatioMode, resetAspectRatioMode,
                                      doc="the widget's pixmap aspect ratio mode")

    #: This property holds the widget's pixmap transformation mode
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QPixmapWidget.getTransformationMode`
    #:     * :meth:`QPixmapWidget.setTransformationMode`
    #:     * :meth:`QPixmapWidget.resetTransformationMode`
    transformationMode = Qt.pyqtProperty("Qt::TransformationMode", getTransformationMode,
                                         setTransformationMode, resetTransformationMode,
                                         doc="the widget's pixmap transformation mode")

    #: This property holds the widget's pixmap alignment
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`QPixmapWidget.getAlignment`
    #:     * :meth:`QPixmapWidget.setAlignment`
    #:     * :meth:`QPixmapWidget.resetAlignment`
    alignment = Qt.pyqtProperty("Qt::Alignment", getAlignment, setAlignment,
                                resetAlignment, doc="the widget's pixmap alignment")


def demo():
    "QPixmap Widget"
    import demo.qpixmapwidgetdemo
    return demo.qpixmapwidgetdemo.main()


def main():
    return demo()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a class which can be used as a super class for all
classes that need to implement the Singleton design pattern."""

__all__ = ["Singleton"]

__docformat__ = "restructuredtext"


class Singleton(object):
    """ This class allows Singleton objects
    The __new__ method is overriden to force Singleton behaviour.
    The Singleton is created for the lowest subclass.
    Usage::

        from taurus.core.util.singleton import Singleton

        class MyManager(Singleton):

            def init(self, *args, **kwargs):
                print "Singleton initialization"

    command line::

        >>> manager1 = MyManager()
        Singleton initialization

        >>> manager2 = MyManager()

        >>> print(manager1,manager2)
        <__main__.MyManager object at 0x9c2a0ec>
        <__main__.MyManager object at 0x9c2a0ec>

    Notice that the two instances of manager point to the same object even
    though you *tried* to construct two instances of MyManager.

    .. warning::

        although __new__ is overriden __init__ is still being called for
        each instance=Singleton()
    """

    # Singleton object
    _the_instance = None

    def __new__(cls, *p, **k):
        # srubio: added type checking
        if cls != type(cls._the_instance):
            cls._the_instance = object.__new__(cls)

            # srubio: added init_single check
            if 'init_single' in cls.__dict__:
                cls._the_instance.init_single(*p, **k)
            else:
                cls._the_instance.init(*p, **k)
        return cls._the_instance

    def init(self, *p, **k):
        pass



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides widgets that display the database in a tree format"""

__all__ = ["TaurusTreeDevicePartItem", "TaurusTreeDeviceDomainItem",
           "TaurusTreeDeviceFamilyItem", "TaurusTreeDeviceMemberItem", "TaurusTreeSimpleDeviceItem",
           "TaurusTreeDeviceItem", "TaurusTreeAttributeItem", "TaurusTreeServerNameItem",
           "TaurusTreeServerItem", "TaurusTreeDeviceClassItem", "TaurusDbBaseModel",
           "TaurusDbSimpleDeviceModel", "TaurusDbPlainDeviceModel", "TaurusDbDeviceModel",
           "TaurusDbSimpleDeviceAliasModel",
           "TaurusDbPlainServerModel", "TaurusDbServerModel",
           "TaurusDbDeviceClassModel",
           "TaurusDbBaseProxyModel",
           "TaurusDbDeviceProxyModel", "TaurusDbServerProxyModel", "TaurusDbDeviceClassProxyModel"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.core.taurusbasetypes import TaurusElementType, TaurusDevState
import taurus.qt.qtcore.mimetypes

from taurus.core.tango.tangodatabase import TangoInfo, TangoDatabase

from .taurusmodel import TaurusBaseTreeItem, TaurusBaseModel, TaurusBaseProxyModel

ElemType = TaurusElementType


def getElementTypeIcon(*args, **kwargs):
    """Wrapper to prevent loading qtgui when this module is imported"""
    from taurus.qt.qtgui.icon import getElementTypeIcon
    return getElementTypeIcon(*args, **kwargs)


def getElementTypeToolTip(*args, **kwargs):
    """Wrapper to prevent loading qtgui when this module is imported"""
    from taurus.qt.qtgui.icon import getElementTypeToolTip
    return getElementTypeToolTip(*args, **kwargs)


def getElementTypeSize(*args, **kwargs):
    """Wrapper to prevent loading qtgui when this module is imported"""
    from taurus.qt.qtgui.icon import getElementTypeSize
    return getElementTypeSize(*args, **kwargs)


def getDevStateIcon(*args, **kwargs):
    """Wrapper to prevent loading qtgui when this module is imported"""
    from taurus.qt.qtgui.icon import getDevStateIcon
    return getDevStateIcon(*args, **kwargs)


def getDevStateToolTip(*args, **kwargs):
    """Wrapper to prevent loading qtgui when this module is imported"""
    from taurus.qt.qtgui.icon import getDevStateToolTip
    return getDevStateToolTip(*args, **kwargs)


class TaurusTreeDbBaseItem(TaurusBaseTreeItem):
    DisplayFunc = TangoInfo.name


class TaurusTreeDevicePartItem(TaurusTreeDbBaseItem):
    """A node designed to represent a 'part' (or totality) of a device name"""

    def data(self, index):
        column = index.column()
        if column > 0:
            return

        model = index.model()
        role = model.role(column, self.depth())

        if role == self.role():
            return self._itemData

    def role(self):
        return ElemType.Device


class TaurusTreeDeviceDomainItem(TaurusTreeDevicePartItem):
    """A node designed to represent a the domain part of a device name"""

    DisplayFunc = str

    def role(self):
        return ElemType.Domain


class TaurusTreeDeviceFamilyItem(TaurusTreeDevicePartItem):
    """A node designed to represent a the family part of a device name"""

    DisplayFunc = str

    def role(self):
        return ElemType.Family


class TaurusTreeDeviceMemberItem(TaurusTreeDevicePartItem):
    """A node designed to represent a the member part of a device name"""

    DisplayFunc = str

    def role(self):
        return ElemType.Member


class TaurusTreeSimpleDeviceItem(TaurusTreeDbBaseItem):
    """A node designed to represent a device (without any child nodes)"""

    def hasChildren(self):
        return False

    def childCount(self):
        return 0

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())
        obj = self.itemData()
        if role == ElemType.Device or role == ElemType.Name:
            return obj.name()
        elif role == ElemType.DeviceAlias:
            return obj.alias()
        elif role == ElemType.Server:
            return obj.server().name()
        elif role == ElemType.DeviceClass:
            return obj.klass().name()
        elif role == ElemType.Exported:
            return obj.state()
        elif role == ElemType.Host:
            return obj.host()
        elif role == ElemType.Domain:
            return obj.domain()
        elif role == ElemType.Family:
            return obj.family()
        elif role == ElemType.Member:
            return obj.member()

    def mimeData(self, index):
        return self.itemData().name()

    def role(self):
        return ElemType.Device


class TaurusTreeDeviceItem(TaurusTreeDbBaseItem):
    """A node designed to represent a device"""

    def child(self, row):
        self.updateChilds()
        return super(TaurusTreeDeviceItem, self).child(row)

    def hasChildren(self):
        return True
        nb = super(TaurusTreeDeviceItem, self).childCount()
        if nb > 0:
            return True
        data = self.itemData()
        if data.state() != TaurusDevState.Ready:
            return False
        return True

    def childCount(self):
        nb = super(TaurusTreeDeviceItem, self).childCount()
        if nb > 0:
            return nb
        data = self.itemData()
        if data.state() != TaurusDevState.Ready:
            return 0
        self.updateChilds()
        return super(TaurusTreeDeviceItem, self).childCount()

    def updateChilds(self):
        if len(self._childItems) > 0:
            return
        for attr in self._itemData.attributes():
            c = TaurusTreeAttributeItem(self._model, attr, self)
            self.appendChild(c)
        return

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())
        obj = self.itemData()
        if role == ElemType.Device or role == ElemType.Name:
            return obj.name()
        elif role == ElemType.DeviceAlias:
            return obj.alias()
        elif role == ElemType.Server:
            return obj.server().name()
        elif role == ElemType.DeviceClass:
            return obj.klass().name()
        elif role == ElemType.Exported:
            return obj.state()
        elif role == ElemType.Host:
            return obj.host()
        elif role == ElemType.Domain:
            return obj.domain()
        elif role == ElemType.Family:
            return obj.family()
        elif role == ElemType.Member:
            return obj.member()

    def mimeData(self, index):
        return self.itemData().fullName()

    def role(self):
        return ElemType.Device


class TaurusTreeAttributeItem(TaurusTreeDbBaseItem):
    """A node designed to represent an attribute"""

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())
        if role == ElemType.Attribute or role == ElemType.Name:
            data = self.itemData()
            data_info = data.info()
            ret = data.name()
            if data_info and hasattr(data_info, 'label'):
                ret = "'" + data_info.label + "' (" + ret + ")"
            return ret

    def toolTip(self, index):
        if index.column() > 0:
            return TaurusTreeDbBaseItem.toolTip(self, index)
        data = self.itemData()
        di = data.info()
        if di is None:
            return data.name()
        ret = '<TABLE border="0" cellpadding="1" cellspacing="0">'
        items = dict(
            name="'" + di.label + "' (" + data.name() + ")",
            description=di.description.replace(
                "<", "&lt;").replace(">", "&gt;"),
            unit=di.unit,
            limits="[%s, %s]" % (di.min_value, di.max_value),
            alarms="[%s, %s]" % (di.alarms.min_alarm, di.alarms.max_alarm),
            warnings="[%s, %s]" % (di.alarms.min_warning, di.alarms.max_warning),)

        for id, value in items.items():
            ret += '<TR><TD WIDTH="80" ALIGN="RIGHT" VALIGN="MIDDLE"><B>%s:</B></TD><TD>%s</TD></TR>' % (
                id.capitalize(), value)
        ret += '</TABLE>'
        return ret

    def mimeData(self, index):
        return self.itemData().fullName()

    def role(self):
        return ElemType.Attribute


class TaurusTreeServerNameItem(TaurusTreeDbBaseItem):
    """A node designed to represent the server name part of a server"""

    DisplayFunc = str

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())

        if role == ElemType.ServerName or role == ElemType.Name:
            return self._itemData

    def role(self):
        return ElemType.ServerName


class TaurusTreeServerItem(TaurusTreeDbBaseItem):
    """A node designed to represent a server"""

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())

        if role == ElemType.Server or role == ElemType.Name:
            return self._itemData.name()
        elif role == ElemType.ServerName:
            return self._itemData.serverName()
        elif role == ElemType.ServerInstance:
            return self._itemData.serverInstance()
        elif role == ElemType.Exported:
            return self._itemData.state()
        elif role == ElemType.Host:
            return self._itemData.host()

    def mimeData(self, index):
        return self.itemData().fullName()

    def role(self):
        return ElemType.Server


class TaurusTreeFullServerItem(TaurusTreeDbBaseItem):
    """A node designed to represent a server"""

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())

        if role == ElemType.Server or role == ElemType.Name:
            return self._itemData.fullName()
        elif role == ElemType.ServerName:
            return self._itemData.serverName()
        elif role == ElemType.ServerInstance:
            return self._itemData.fullName()
        elif role == ElemType.Exported:
            return self._itemData.state()
        elif role == ElemType.Host:
            return self._itemData.host()

    def mimeData(self, index):
        return self.itemData().fullName()

    def role(self):
        return ElemType.Server


class TaurusTreeDeviceClassItem(TaurusTreeDbBaseItem):
    """A node designed to represent a device class"""

    def data(self, index):
        column, model = index.column(), index.model()
        role = model.role(column, self.depth())

        if role == ElemType.Name or role == ElemType.DeviceClass:
            return self._itemData.name()

    def mimeData(self, index):
        return self.itemData().fullName()

    def role(self):
        return ElemType.DeviceClass


class TaurusDbBaseModel(TaurusBaseModel):
    """The base class for all Taurus database Qt models.
    By default, this model represents a plain device perspective of the underlying
    database."""

    ColumnNames = "Device", "Alias", "Server", "Class", "Alive", "Host"
    ColumnRoles = (
        ElemType.Device, ElemType.Device), ElemType.DeviceAlias, ElemType.Server, ElemType.DeviceClass, ElemType.Exported, ElemType.Host

    def createNewRootItem(self):
        return TaurusTreeDbBaseItem(self, self.ColumnNames)

    def refresh(self, refresh_source=False):
        data = self.dataSource()
        if refresh_source and data is not None:
            data.refreshCache()
        TaurusBaseModel.refresh(self, refresh_source=refresh_source)

    def roleIcon(self, taurus_role):
        return getElementTypeIcon(taurus_role)

    def columnIcon(self, column):
        return self.roleIcon(self.role(column))

    def roleToolTip(self, taurus_role):
        return getElementTypeToolTip(taurus_role)

    def columnToolTip(self, column):
        return self.roleToolTip(self.role(column))

    def roleSize(self, taurus_role):
        return getElementTypeSize(taurus_role)

    def columnSize(self, column):
        taurus_role = self.role(column)
        s = self.roleSize(taurus_role)
        return s

    def mimeTypes(self):
        return ["text/plain", taurus.qt.qtcore.mimetypes.TAURUS_MODEL_LIST_MIME_TYPE, taurus.qt.qtcore.mimetypes.TAURUS_MODEL_MIME_TYPE]

    def mimeData(self, indexes):
        ret = Qt.QMimeData()
        data = []
        for index in indexes:
            if not index.isValid():
                continue
            #item = index.internalPointer()
            #data = item.mimeData(index)
            #stream << item.mimeData(index)
            tree_item = index.internalPointer()
            mime_data_item = tree_item.mimeData(index)
            if mime_data_item is None:
                continue
            data.append(mime_data_item)
        ret.setData(
            taurus.qt.qtcore.mimetypes.TAURUS_MODEL_LIST_MIME_TYPE, "\r\n".join(data))
        ret.setText(", ".join(data))
        if len(data) == 1:
            ret.setData(
                taurus.qt.qtcore.mimetypes.TAURUS_MODEL_MIME_TYPE, str(data[0]))
        return ret

    def pyData(self, index, role):
        if not index.isValid():
            return None

        item = index.internalPointer()
        row, column, depth = index.row(), index.column(), item.depth()
        taurus_role = self.role(column, depth)

        ret = None
        if role == Qt.Qt.DisplayRole:
            if taurus_role != ElemType.Exported:
                ret = item.data(index)
        elif role == Qt.Qt.DecorationRole:
            if column == 0:
                ret = self.roleIcon(taurus_role)
            if taurus_role == ElemType.Exported:
                state = item.data(index)
                ret = getDevStateIcon(state)
        elif role == Qt.Qt.ToolTipRole:
            ret = item.toolTip(index)
            if ret is None:
                data = item.data(index)
                if data is not None:
                    if taurus_role == ElemType.Exported:
                        ret = getDevStateToolTip(data)
                    else:
                        ret = self.roleToolTip(taurus_role)
        # elif role == Qt.Qt.SizeHintRole:
        #    ret = self.columnSize(column)
        elif role == Qt.Qt.FontRole:
            ret = self.DftFont
        return ret

    def setupModelData(self, data):
        if data is None:
            return
        if isinstance(data, TangoDatabase):
            data = data.cache()
        devices = data.devices()

        rootItem = self._rootItem
        for dev_name in data.getDeviceNames():
            dev = devices[dev_name]
            devItem = TaurusTreeSimpleDeviceItem(self, dev, rootItem)
            rootItem.appendChild(devItem)


class TaurusDbSimpleDeviceModel(TaurusDbBaseModel):
    """A Qt model that structures device elements in 1 level tree with
    device name as node leafs. This model contains only 1 column."""

    ColumnNames = "Device",
    ColumnRoles = (ElemType.Device, ElemType.Device),


class TaurusDbSimpleDeviceAliasModel(TaurusDbBaseModel):
    """A Qt model that structures device elements in 1 level tree with
    device alias as node leafs. This model contains only 1 column."""

    ColumnNames = "Alias",
    ColumnRoles = (ElemType.DeviceAlias, ElemType.DeviceAlias),

    def setupModelData(self, data):
        if data is None:
            return
        if isinstance(data, TangoDatabase):
            data = data.cache()
        devices = data.devices()

        rootItem = self._rootItem
        for dev_name in data.getDeviceNames():
            dev = devices[dev_name]
            if dev.alias() is not None:
                devItem = TaurusTreeSimpleDeviceItem(self, dev, rootItem)
                rootItem.appendChild(devItem)


class TaurusDbPlainDeviceModel(TaurusDbBaseModel):
    """A Qt model that structures device elements in 1 level tree. Device
    nodes will have attribute child nodes if the device is running."""

    ColumnNames = "Device", "Alias", "Server", "Class", "Alive", "Host"
    ColumnRoles = (ElemType.Device, ElemType.Device,
                   ElemType.Attribute), ElemType.DeviceAlias, ElemType.Server, ElemType.DeviceClass, ElemType.Exported, ElemType.Host

    def setupModelData(self, data):
        if data is None:
            return
        if isinstance(data, TangoDatabase):
            data = data.cache()
        devices = data.devices()

        rootItem = self._rootItem
        for dev_name in data.getDeviceNames():
            dev = devices[dev_name]
            devItem = TaurusTreeDeviceItem(self, dev, rootItem)
            rootItem.appendChild(devItem)


class TaurusDbDeviceModel(TaurusDbBaseModel):
    """A Qt model that structures device elements is a 3 level tree organized
       as:

           - <domain>
           - <family>
           - <member>"""
    ColumnRoles = (ElemType.Device, ElemType.Domain, ElemType.Family, ElemType.Member,
                   ElemType.Attribute), ElemType.DeviceAlias, ElemType.Server, ElemType.DeviceClass, ElemType.Exported, ElemType.Host

    def setupModelData(self, data):
        if data is None:
            return
        if isinstance(data, TangoDatabase):
            data = data.deviceTree()

        rootItem = self._rootItem
        for domain in data.keys():
            families = data[domain]
            domainItem = TaurusTreeDeviceDomainItem(
                self, domain.upper(), rootItem)
            for family in families.keys():
                members = families[family]
                familyItem = TaurusTreeDeviceFamilyItem(
                    self, family.upper(), domainItem)
                for member in members.keys():
                    dev = members[member]
                    memberItem = TaurusTreeDeviceItem(
                        self, dev, parent=familyItem)
                    familyItem.appendChild(memberItem)
                domainItem.appendChild(familyItem)
            rootItem.appendChild(domainItem)


class TaurusDbPlainServerModel(TaurusDbBaseModel):
    ColumnNames = "Server", "Alive", "Host"
    ColumnRoles = (ElemType.Server,
                   ElemType.ServerInstance), ElemType.Exported, ElemType.Host

    def setupModelData(self, data):
        if data is None:
            return

        if isinstance(data, TangoDatabase):
            data = data.cache()

        servers = data.servers()
        rootItem = self._rootItem

        for server_name, server in servers.items():
            serverInstanceItem = TaurusTreeFullServerItem(
                self, server, rootItem)
            rootItem.appendChild(serverInstanceItem)


class TaurusDbServerModel(TaurusDbBaseModel):
    """A Qt model that structures server elements in a tree organized
    as:

        - <Server name>
        - <Server instance>
        - <Class>
        - <Device>
        - <Attribute>"""

    ColumnNames = "Server", "Alive", "Host"
    ColumnRoles = (ElemType.Server, ElemType.ServerName, ElemType.ServerInstance,
                   ElemType.DeviceClass, ElemType.Device, ElemType.Attribute), ElemType.Exported, ElemType.Host

    def setupModelData(self, data):
        if data is None:
            return

        if isinstance(data, TangoDatabase):
            data = data.cache()

        servers, klasses, devices = data.servers(), data.klasses(), data.devices()
        rootItem = self._rootItem
        server_dict = {}

        server_names = data.getServerNames()
        for server_name in server_names:
            server = servers[server_name]
            name, instance = server.serverName(), server.serverInstance()

            serverNameItem = server_dict.get(name)
            if serverNameItem is None:
                serverNameItem = TaurusTreeServerNameItem(self, name, rootItem)
                rootItem.appendChild(serverNameItem)
                server_dict[name] = serverNameItem
                # rootItem.appendChild(serverNameItem)

            serverInstanceItem = TaurusTreeServerItem(
                self, server, serverNameItem)
            serverNameItem.appendChild(serverInstanceItem)

            klass_names = server.getClassNames()
            device_names = server.getDeviceNames()
            for klass_name in klass_names:
                klass = klasses[klass_name]
                klassItem = TaurusTreeDeviceClassItem(
                    self, klass, serverInstanceItem)
                serverInstanceItem.appendChild(klassItem)
                for dev_name in device_names:
                    dev = devices[dev_name]
                    if dev.klass() == klass:
                        devItem = TaurusTreeDeviceItem(self, dev, klassItem)
                        klassItem.appendChild(devItem)


class TaurusDbDeviceClassModel(TaurusDbBaseModel):
    """A Qt model that structures class elements in a tree organized as:

       * <Class>
           * <Device>
               * <Attribute>"""
    ColumnNames = "Class", "Alive", "Host"
    ColumnRoles = (ElemType.DeviceClass, ElemType.DeviceClass,
                   ElemType.Device, ElemType.Attribute), ElemType.Exported, ElemType.Host

    def setupModelData(self, data):
        if data is None:
            return

        if isinstance(data, TangoDatabase):
            data = data.cache()

        rootItem = self._rootItem
        klasses, devices = data.klasses(), data.devices()
        dev_nb = 0
        for klass_name in data.getClassNames():
            klass = klasses[klass_name]
            klassItem = TaurusTreeDeviceClassItem(self, klass, rootItem)
            for dev_name in klass.getDeviceNames():
                dev_nb += 1
                dev = devices[dev_name]
                devItem = TaurusTreeDeviceItem(self, dev, klassItem)
                klassItem.appendChild(devItem)
            rootItem.appendChild(klassItem)


class TaurusDbBaseProxyModel(TaurusBaseProxyModel):
    pass


class TaurusDbDeviceProxyModel(TaurusDbBaseProxyModel):
    """A Qt filter & sort model for model for the taurus models:
           - TaurusDbBaseModel
           - TaurusDbDeviceModel
           - TaurusDbSimpleDeviceModel
           - TaurusDbPlainDeviceModel"""

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()
        regexp = self.filterRegExp()

        # if domain node, check if it will potentially have any children
        if isinstance(treeItem, TaurusTreeDeviceDomainItem):
            domain = treeItem.display()
            devices = sourceModel.getDomainDevices(domain)
            for device in devices:
                if self.deviceMatches(device, regexp):
                    return True
            return False

        # if family node, check if it will potentially have any children
        if isinstance(treeItem, TaurusTreeDeviceFamilyItem):
            domain = treeItem.parent().display()
            family = treeItem.display()
            devices = sourceModel.getFamilyDevices(domain, family)
            for device in devices:
                if self.deviceMatches(device, regexp):
                    return True
            return False

        if isinstance(treeItem, TaurusTreeDeviceItem) or \
           isinstance(treeItem, TaurusTreeSimpleDeviceItem) or \
           isinstance(treeItem, TaurusTreeDeviceMemberItem):
            device = treeItem.itemData()
            return self.deviceMatches(device, regexp)
        return True

    def deviceMatches(self, device, regexp):
        name = device.name()

        # if Qt.QString(name).contains(regexp):
        if regexp.indexIn(name) != -1:
            return True
        name = device.alias()
        if name is None:
            return False
        # return Qt.QString(name).contains(regexp)
        return regexp.indexIn(name) != -1


class TaurusDbServerProxyModel(TaurusDbBaseProxyModel):
    """A Qt filter & sort model for the TaurusDbServerModel"""

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()
        regexp = self.filterRegExp()

        # if server name node, check if it will potentially have any children
        if isinstance(treeItem, TaurusTreeServerNameItem):
            serverName = treeItem.display()
            serverInstances = sourceModel.getServerNameInstances(serverName)
            for serverInstance in serverInstances:
                # if Qt.QString(serverInstance.name()).contains(regexp):
                if regexp.indexIn(serverInstance.name()) != -1:
                    return True
            return False

        if isinstance(treeItem, TaurusTreeServerItem):
            # return treeItem.qdisplay().contains(regexp)
            return regexp.indexIn(treeItem.qdisplay()) != -1

        return True


class TaurusDbDeviceClassProxyModel(TaurusDbBaseProxyModel):
    """A Qt filter & sort model for the TaurusDbDeviceClassModel"""

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()

        if not isinstance(treeItem, TaurusTreeDeviceClassItem):
            return True

        regexp = self.filterRegExp()

        # return treeItem.qdisplay().contains(regexp)
        return regexp.indexIn(treeItem.qdisplay()) != -1



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
"""
taurusgraphic.py:
"""

__all__ = ['SynopticSelectionStyle',
           'parseTangoUri',
           'QEmitter',
           'TaurusGraphicsUpdateThread',
           'TaurusGraphicsScene',
           'QGraphicsTextBoxing',
           'QSpline',
           'TaurusGraphicsItem',
           'TaurusGraphicsAttributeItem',
           'TaurusGraphicsStateItem',
           'TaurusEllipseStateItem',
           'TaurusRectStateItem',
           'TaurusSplineStateItem',
           'TaurusRoundRectItem',
           'TaurusRoundRectStateItem',
           'TaurusGroupItem',
           'TaurusGroupStateItem',
           'TaurusPolygonStateItem',
           'TaurusLineStateItem',
           'TaurusTextStateItem',
           'TaurusTextAttributeItem',
           'TYPE_TO_GRAPHICS',
           'TaurusBaseGraphicsFactory',
           ]

__docformat__ = 'restructuredtext'

import re
import os
import subprocess
import traceback
import operator
import types

import Queue

from taurus import Manager
from taurus.core import AttrQuality, DataType
from taurus.core.util.containers import CaselessDefaultDict
from taurus.core.util.log import Logger
from taurus.core.taurusdevice import TaurusDevice
from taurus.core.taurusattribute import TaurusAttribute
from taurus.core.util.enumeration import Enumeration
# TODO: tango-centric!
from taurus.core.tango import DevState
from taurus.core.tango.tangovalidator import (TangoDeviceNameValidator,
                                              TangoAttributeNameValidator)
from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseComponent
from taurus.qt.qtgui.util import (QT_ATTRIBUTE_QUALITY_PALETTE, QT_DEVICE_STATE_PALETTE,
                                  ExternalAppAction, TaurusWidgetFactory)


SynopticSelectionStyle = Enumeration("SynopticSelectionStyle", [
    # A blue ellipse is displayed around the selected objects
    "ELLIPSE",
    # The own outline of selected object is displayed in blue and bolder
    "OUTLINE",
])


def parseTangoUri(name):
    from taurus.core import tango
    validator = {tango.TangoDevice: TangoDeviceNameValidator,
                 tango.TangoAttribute: TangoAttributeNameValidator}
    try:
        val = validator[tango.TangoFactory().findObjectClass(name)]()
        params = val.getUriGroups(name)
        return (params if '_devslashname' in params else None)
    except:
        return None


class QEmitter(Qt.QObject):
    updateView = Qt.pyqtSignal('QGraphicsView')


class TaurusGraphicsUpdateThread(Qt.QThread):

    def __init__(self, parent=None, period=3):
        """Parent most not be None and must be a TaurusGraphicsScene!"""
        if not isinstance(parent, TaurusGraphicsScene):
            raise RuntimeError("Illegal parent for TaurusGraphicsUpdateThread")
        Qt.QThread.__init__(self, parent)
        self.period = period
        self.log = Logger('TaurusGraphicsUpdateThread')

    def _updateView(self, v):
        # The first one is the prefered one because it improves performance
        # since updates don't come very often in comparison to with the refresh
        # rate of the monitor (~70Hz)
        if v.viewportUpdateMode() == Qt.QGraphicsView.NoViewportUpdate:
            # We call the update to the viewport instead of the view
            # itself because apparently there is a bug in QT 4.3 that
            # prevents a proper update when the view is inside a QTab
            v.viewport().update()
        else:
            # @todo This is probably a bug (item_rects is not defined). But it is defined in .run(), see "todo" below...
            v.updateScene(item_rects)
            # v.invalidateScene(item.boundingRect())
        return

    def run(self):
        self.log.debug("run... - TaurusGraphicsUpdateThread")
        emitter = QEmitter()
        emitter.moveToThread(Qt.QApplication.instance().thread())
        emitter.setParent(Qt.QApplication.instance())
        emitter.updateView.connect(self._updateView)

        p = self.parent()
        while True:
            item = p.getQueue().get(True)
            if type(item) in types.StringTypes:
                if item == "exit":
                    break
                else:
                    continue
            if not operator.isSequenceType(item):
                item = (item,)
            # @todo: Unless the call to boundingRect() has a side effect, this line is useless..  probably related to todo in _updateView()
            item_rects = [i.boundingRect() for i in item]

            for v in p.views():
                # p.debug("emit('updateView')")
                # emitter.updateView.emit(v)
                self.emitter.updateView.emit(v)
            # This sleep is needed to reduce CPU usage of the application!
            self.sleep(self.period)
            # End of while
        # End of Thread


class TaurusGraphicsScene(Qt.QGraphicsScene):
    '''
    This class encapsulates TaurusJDrawSynopticsView and TaurusGraphicsScene signals/slots

    External events::

     Slot selectGraphicItem(const QString &) displays a selection
     mark around the TaurusGraphicsItem that matches the argument passed.

    Mouse Left-button events::

     Signal graphicItemSelected(QString) is triggered, passing the
     selected TaurusGraphicsItem.name() as argument.

    Mouse Right-button events::

     TaurusGraphicsItem.setContextMenu([(ActionName,ActionMethod(device_name))]
     allows to configure custom context menus for graphic items using a list
     of tuples. Empty tuples will insert separators in the menu.
    '''
    ANY_ATTRIBUTE_SELECTS_DEVICE = True
    TRACE_ALL = False

    refreshTree2 = Qt.pyqtSignal()
    graphicItemSelected = Qt.pyqtSignal('QString')
    graphicSceneClicked = Qt.pyqtSignal('QPoint')

    def __init__(self, parent=None, strt=True):
        name = self.__class__.__name__
        # self.call__init__(Logger, name, parent) #Inheriting from Logger
        # caused exceptions in CONNECT
        Qt.QGraphicsScene.__init__(self, parent)
        self.updateQueue = None
        self.updateThread = None
        self._itemnames = CaselessDefaultDict(lambda k: set())
        self._selection = []
        self._selectedItems = []
        self._selectionStyle = SynopticSelectionStyle.OUTLINE
        self.threads = []
        self.pids = []
        self.panels = []
        self.panel_launcher = None

        try:
            self.logger = Logger(name)
            # self.logger.setLogLevel(self.logger.Info)
            if not self.TRACE_ALL:
                self.debug = lambda l: self.logger.debug(l)
                self.info = lambda l: self.logger.info(l)
                self.warning = lambda l: self.logger.warning(l)
            else:
                self.debug = self.info = self.warning = self.error = lambda l: self.logger.warning(
                    l)
        except:
            print 'Unable to initialize TaurusGraphicsSceneLogger: %s' % traceback.format_exc()

        try:
            if parent and parent.panelClass() is not None:
                defaultClass = parent.panelClass()
                if defaultClass and isinstance(defaultClass, str):
                    self.panel_launcher = self.getClass(defaultClass)
                    if self.panel_launcher is None:
                        self.panel_launcher = ExternalAppAction(
                            defaultClass.split())
                else:
                    self.panel_launcher = defaultClass
            else:
                from taurus.qt.qtgui.graphic.jdraw import TaurusJDrawSynopticsView
                self.panel_launcher = TaurusJDrawSynopticsView.defaultPanelClass()
        except:
            self.warning(traceback.format_exc())
            self.panel_launcher = None

        self.setSelectionMark()
        if strt:
            self.start()

    def __del__(self):
        self.closeAllPanels()
        Qt.QGraphicsScene.__del__(self)

    def showNewPanel(self, args=None, standAlone=False):
        try:
            if isinstance(args, TaurusGraphicsItem):
                objName = args._name
                clName = args.getExtensions().get('className') or self.panel_launcher
                # classParams extension overrides Model; if there's no
                # extension then object name is used
                clParam = args.getExtensions().get('classParams') or objName
                standAlone = args.standAlone
            else:
                clName, clParam, objName = self.panel_launcher, args, args
            if not clName or clName == 'noPanel':
                return
            self.debug('TaurusGraphicsScene.showNewPanel(%s,%s,%s)' %
                       (clName, clParam, objName))
            if isinstance(clName, ExternalAppAction):
                clName.actionTriggered(clParam if isinstance(
                    clParam, (list, tuple)) else [clParam])
            else:
                if isinstance(clName, str):
                    klass = self.getClass(clName)
                    if klass is None:
                        self.warning("%s Class not found!" % clName)
                        klass = self.getClass("TaurusDevicePanel")
                else:
                    klass, clName = clName, getattr(
                        clName, '__name__', str(clName))
                widget = klass()  # self.parent())
                # if isinstance(widget,taurus.qt.qtgui.panel.TaurusDevicePanel):
                # widget.setSpectraAtkMode(True) #Method renamed or deprecated
                try:
                    widget.setClasses(clParam)
                except:
                    pass
                try:
                    widget.setModel(clParam)
                except:
                    pass
                try:
                    widget.setTable(clParam)
                except:
                    pass

                # if isinstance(widget,Qt.QWidget):
                    # if not standAlone:
                        #obj = newDialog(self.parent())
                    # else:
                        #obj = newDialog()
                    # obj.initComponents(widget,objName,clName)
                    # obj.setModal(False)
                    # obj.setVisible(True)

                widget.setWindowTitle('%s - %s' % (clName, objName))
                self.panels.append(widget)
                widget.show()  # exec_()
                return widget
        except:
            self.warning(traceback.format_exc())

    def closeAllPanels(self):
        """ This method replaces killProcess, using taurus.qt.qtgui.util.ExternalAppAction instead! """
        try:
            self.debug('In closeAllPanels(%s,%s)' %
                       (self.panel_launcher, self.panels))
            if isinstance(self.panel_launcher, ExternalAppAction):
                self.panel_launcher.kill()
            for p in self.panels:
                try:
                    if hasattr(p, 'setModel'):
                        p.setModel(None)
                    p.close()
                except:
                    pass
            while self.panels:
                self.panels.pop(0)
        except:
            self.warning(traceback.format_exc())

    def addItem(self, item):
        # self.debug('addItem(%s)'%item)
        def expand(i):
            name = str(getattr(i, '_name', '')).lower()
            if name:
                self._itemnames[name].add(i)
                #self.debug('addItem(%s): %s'%(name,i))
            if isinstance(i, Qt.QGraphicsItemGroup):
                for j in i.childItems():
                    expand(j)
        expand(item)
        Qt.QGraphicsScene.addItem(self, item)

    def addWidget(self, item, flags=None):
        self.debug('addWidget(%s)' % item)
        name = str(getattr(item, '_name', '')).lower()
        if name:
            self._itemnames[name].add(item)
        if flags is None:
            Qt.QGraphicsScene.addWidget(self, item)
        else:
            Qt.QGraphicsScene.addWidget(self, item, flags)

    def getItemByName(self, item_name, strict=None):
        """
        Returns a list with all items matching a given name.

        :param strict: (bool or None) controls whether full_name (strict=True) or only device name (False) must match

        :return: (list) items
        """
        strict = (
            not self.ANY_ATTRIBUTE_SELECTS_DEVICE) if strict is None else strict
        alnum = '(?:[a-zA-Z0-9-_\*]|(?:\.\*))(?:[a-zA-Z0-9-_\*]|(?:\.\*))*'
        target = str(item_name).strip().split()[0].lower().replace(
            '/state', '')  # If it has spaces only the first word is used
        # Device names should match also its attributes or only state?
        if not strict and TangoAttributeNameValidator().getUriGroups(target):
            target = target.rsplit('/', 1)[0]
        if TangoDeviceNameValidator().getUriGroups(target):
            if strict:
                target += '(/state)?'
            else:
                target += '(/' + alnum + ')?'
        if not target.endswith('$'):
            target += '$'
        result = []
        for k in self._itemnames.keys():
            if re.match(target.lower(), k.lower()):
                #self.debug('getItemByName(%s): _itemnames[%s]: %s'%(target,k,self._itemnames[k]))
                result.extend(self._itemnames[k])
        return result

    def getItemByPosition(self, x, y):
        """ This method will try first with named objects; if failed then with itemAt """
        pos = Qt.QPointF(x, y)
        itemsAtPos = []
        for z, o in sorted((i.zValue(), i) for v in self._itemnames.values() for i in v if i.contains(pos) or i.isUnderMouse()):
            if not hasattr(o, 'getExtensions'):
                self.debug(
                    'getItemByPosition(%d,%d): adding Qt primitive %s' % (x, y, o))
                itemsAtPos.append(o)
            elif not o.getExtensions().get('noSelect'):
                self.debug(
                    'getItemByPosition(%d,%d): adding GraphicsItem %s' % (x, y, o))
                itemsAtPos.append(o)
            else:
                self.debug(
                    'getItemByPosition(%d,%d): object ignored, %s' % (x, y, o))
        if itemsAtPos:
            obj = itemsAtPos[-1]
            return self.getTaurusParentItem(obj) or obj
        else:
            # return self.itemAt(x,y)
            self.debug('getItemByPosition(%d,%d): no items found!' % (x, y))
            return None

    def getItemClicked(self, mouseEvent):
        pos = mouseEvent.scenePos()
        x, y = pos.x(), pos.y()
        self.graphicSceneClicked.emit(Qt.QPoint(x, y))
        obj = self.getItemByPosition(x, y)
        #self.debug('mouse clicked on %s(%s) at (%s,%s)'%(type(obj).__name__,getattr(obj,'_name',''),x,y))
        return obj

    def mousePressEvent(self, mouseEvent):
        #self.debug('In TaurusGraphicsScene.mousePressEvent(%s,%s))'%(str(type(mouseEvent)),str(mouseEvent.button())))
        try:
            obj = self.getItemClicked(mouseEvent)
            obj_name = getattr(obj, '_name', '')
            if not obj_name and isinstance(obj, QGraphicsTextBoxing):
                obj_name = obj.toPlainText()
            if (mouseEvent.button() == Qt.Qt.LeftButton):
                # A null obj_name should deselect all, we don't send obj
                # because we want all similar to be matched
                if self.selectGraphicItem(obj_name):
                    self.debug(
                        ' => graphicItemSelected(QString)(%s)' % obj_name)
                    self.graphicItemSelected.emit(obj_name)
                else:
                    # It should send None but the signature do not allow it
                    self.graphicItemSelected.emit("")

            def addMenuAction(menu, k, action, last_was_separator=False):
                try:
                    if k:
                        configDialogAction = menu.addAction(k)
                        if action:
                            configDialogAction.triggered[()].connect(lambda dev=obj_name, act=action: act(dev))
                        else:
                            configDialogAction.setEnabled(False)
                        last_was_separator = False
                    elif not last_was_separator:
                        menu.addSeparator()
                        last_was_separator = True
                except Exception, e:
                    self.warning('Unable to add Menu Action: %s:%s' % (k, e))
                return last_was_separator
            if (mouseEvent.button() == Qt.Qt.RightButton):
                ''' This function is called when right clicking on TaurusDevTree area. A pop up menu will be shown with the available options. '''
                self.debug('RightButton Mouse Event on %s' % (obj_name))
                if isinstance(obj, TaurusGraphicsItem) and (obj_name or obj.contextMenu() or obj.getExtensions()):
                    menu = Qt.QMenu(None)  # self.parent)
                    last_was_separator = False
                    extensions = obj.getExtensions()
                    if obj_name and (not extensions or not extensions.get('className')):
                        # menu.addAction(obj_name)
                        addMenuAction(menu, 'Show %s panel' % obj_name,
                                      lambda x=obj_name: self.showNewPanel(x))
                    if obj.contextMenu():
                        if obj_name:
                            menu.addSeparator()
                            last_was_separator = True
                        for t in obj.contextMenu():  # It must be a list of tuples (ActionName,ActionMethod)
                            last_was_separator = addMenuAction(
                                menu, t[0], t[1], last_was_separator)
                    if extensions:
                        if not menu.isEmpty():
                            menu.addSeparator()
                        className = extensions.get('className')
                        if className and className != 'noPanel':
                            self.debug('launching className extension object')
                            addMenuAction(menu, 'Show %s' % className,
                                          lambda d, x=obj: self.showNewPanel(x))
                        if extensions.get('shellCommand'):
                            addMenuAction(menu, 'Execute', lambda d,
                                          x=obj: self.getShellCommand(x))
                    if not menu.isEmpty():
                        menu.exec_(
                            Qt.QPoint(mouseEvent.screenPos().x(), mouseEvent.screenPos().y()))
                    del menu
        except Exception:
            self.warning(traceback.format_exc())

    def mouseDoubleClickEvent(self, event):
        try:
            obj = self.getItemClicked(event)
            obj_name = getattr(obj, '_name', '')
            try:
                class_name = obj.getExtensions().get('className')
            except:
                class_name = 'noPanel'
            self.debug('Clicked (%s,%s,%s)' % (obj, obj_name, class_name))
            if obj_name and class_name != 'noPanel':
                self.showNewPanel(obj)
        except:
            self.warning(traceback.format_exc())

    def setSelectionStyle(self, selectionStyle):
        # TODO We should test that selectionStyle is part of
        # SynopticSelectionStyle but there is nothing about it
        self._selectionStyle = selectionStyle

    def selectGraphicItem(self, item_name):
        """
        A blue circle is drawn around the matching item name.
        If the item_name is empty, or it is a reserved keyword, or it has the "noSelect" extension, then the blue circle is removed from the synoptic.
        """
        #self.debug('In TaurusGraphicsScene.selectGraphicItem(%s))'%item_name)
        selected = [str(getattr(item, '_name', item))
                    for item in self._selectedItems if item]
        if selected:
            iname = str(getattr(item_name, '_name', item_name))
            #self.debug('In TauGraphicsScene.selectGraphicItem(%s): already selected: %s'%(iname,selected))
            if not iname.strip():
                self.clearSelection()
                return False
            elif any(iname not in i for i in selected):
                self.clearSelection()
            else:
                self.debug(
                    'In TauGraphicsScene.selectGraphicItem(%s): already selected!' % item_name)
                return True
        if any(isinstance(item_name, t) for t in (TaurusGraphicsItem, Qt.QGraphicsItem)):
            if not getattr(item_name, '_name', ''):
                self.debug(
                    'In TauGraphicsScene.selectGraphicItem(%s): item name not found.' % item_name)
                return False
            items = [item_name]
        else:
            from taurus.qt.qtgui.graphic import jdraw_parser
            if not item_name or (str(item_name).startswith('JD') and str(item_name) in jdraw_parser.reserved):
                self.debug(
                    'In TauGraphicsScene.selectGraphicItem(%s): item name not found or name is a reserved keyword.' % item_name)
                return False
            items = self.getItemByName(item_name) or []
            items = [i for i in items if self.getTaurusParentItem(
                i) not in (items + self._selectedItems)]
            self.debug('In TaurusGraphicsScene.selectGraphicItem(%s)): matched %d items' % (
                item_name, len(items)))

        if self._selectionStyle == SynopticSelectionStyle.ELLIPSE:
            displaySelection = self._displaySelectionAsEllipse
        elif self._selectionStyle == SynopticSelectionStyle.OUTLINE:
            displaySelection = self._displaySelectionAsOutline
        else:
            raise Exception("Unexpected selectionStyle '%s'" %
                            SynopticSelectionStyle.whatis(self._selectionStyle))
        return displaySelection(items)

    def _displaySelectionAsEllipse(self, items):
        retval = False
        for item in items:
            try:
                if ((isinstance(item, TaurusGraphicsItem) and item.getExtensions().get('noSelect'))
                            or (item in self._selection)
                            # or (item in tangoGroup)
                        ):
                    continue
                x, y = item.x(), item.y()
                rect = item.boundingRect()
                srect = self.sceneRect()
                # 0 has to be excluded to check grouped element
                if not (0 < x <= self.sceneRect().width() and 0 < y <= srect.height()):
                    rx, ry = rect.topLeft().x(), rect.topLeft().y()
                    self.debug('\tposition not well mapped (%s,%s), using rect bound (%s,%s) instead' % (
                        x, y, rx, ry))
                    x, y = rx, ry  # If the object is in the corner it will be also 0
                w, h = rect.width(), rect.height()
                if x < 0 or y < 0:
                    self.debug('Cannot draw SelectionMark for %s(%s)(%s,%s) in a negative position (%f,%f)' % (
                        type(item).__name__, item._name, w, h, x, y))
                else:
                    if type(item) in (TaurusTextAttributeItem, TaurusTextStateItem) and isinstance(self.getSelectionMark(), Qt.QGraphicsPixmapItem):
                        x, y, w, h = x - 20, y, 20, 20
                    self.drawSelectionMark(x, y, w, h)
                    self.debug('> Moved the SelectionMark to item %s(%s)(%s,%s) at %f,%f' % (
                        type(item).__name__, item._name, w, h, x, y))
                if item not in self._selectedItems:
                    self._selectedItems.append(item)
                retval = True
            except Exception, e:
                self.warning('selectGraphicsItem(%s) failed! %s' %
                             (getattr(item, '_name', item), str(e)))
                self.warning(traceback.format_exc())
                # return False
        return retval

    def _displaySelectionAsOutline(self, items):
        def _outline(shapes):
            """"Compute the boolean union from a list of QGraphicsItem. """
            shape = None
            # TODO we can use a stack instead of recursivity
            for s in shapes:
                # TODO we should skip text and things like that
                if isinstance(s, TaurusGroupItem):
                    s = _outline(s.childItems())
                    if s is None:
                        continue

                s = s.shape()
                if shape is not None:
                    shape = shape.united(s)
                else:
                    shape = s

            if shape is None:
                return None

            return Qt.QGraphicsPathItem(shape)

        # TODO we can cache the outline instead of computing it again and again
        selectionShape = _outline(items)
        if selectionShape:
            # copy-paste from getSelectionMark
            color = Qt.QColor(Qt.Qt.blue)
            color.setAlphaF(.10)
            pen = Qt.QPen(Qt.Qt.SolidLine)
            pen.setWidth(4)
            pen.setColor(Qt.QColor(Qt.Qt.blue))
            selectionShape.setBrush(color)
            selectionShape.setPen(pen)

            for item in items:
                if item not in self._selectedItems:
                    self._selectedItems.append(item)

            # TODO i dont think this function work... or i dont know how...
            # self.setSelectionMark(picture=selectionShape)
            # ... Then do it it with hands...
            # copy-paste from drawSelectionMark
            self._selection.append(selectionShape)
            # It's better to add it hidden to avoid resizings
            selectionShape.hide()
            self.addItem(selectionShape)
            # Put on Top
            selectionShape.setZValue(9999)
            selectionShape.show()
            self.updateSceneViews()

            return True

        return False

    def clearSelection(self):
        #self.debug('In clearSelection([%d])'%len(self._selectedItems))
        for i in self._selection:
            i.hide()
            self.removeItem(i)
        self._selection = []
        self._selectedItems = []
        self.updateSceneViews()

    def setSelectionMark(self, picture=None, w=10, h=10):
        """This method allows to set a callable, graphic item or pixmap as selection mark (by default creates a blue circle).
        If picture is a callable, the object returned will be used as selection mark.
        If picture is a QGraphicsItem it will be used as selection mark.
        If picture is a QPixmap or a path to a pixmap a QGraphicsPixmapItem will be created.
        If no picture is provided, a blue ellipse will be drawn around the selected object.
        h/w will be used for height/width of the drawn object.
        """
        #self.debug('In setSelectionMark(%s,%d,%d)'%(picture,w,h))
        if picture is None:
            self.SelectionMark = None  # Reset of previous icon generators
        else:
            self.SelectionMark = (lambda p=picture, x=w,
                                  y=h: self.getSelectionMark(p, x, y))
        return self.SelectionMark

    def getSelectionMark(self, picture=None, w=10, h=10):
        if picture is None:
            if self.SelectionMark:
                SelectionMark = self.SelectionMark()
            else:
                SelectionMark = Qt.QGraphicsEllipseItem()
                color = Qt.QColor(Qt.Qt.blue)
                color.setAlphaF(.10)
                SelectionMark.setBrush(color)
                pen = Qt.QPen(Qt.Qt.CustomDashLine)
                pen.setWidth(4)
                pen.setColor(Qt.QColor(Qt.Qt.blue))
                SelectionMark.setPen(pen)
                SelectionMark.hide()  # It's better to add it hidden to avoid resizings
        else:
            try:
                if isinstance(picture, Qt.QGraphicsItem):
                    SelectionMark = picture
                    SelectionMark.setRect(0, 0, w, h)
                    SelectionMark.hide()
                elif operator.isCallable(picture):
                    SelectionMark = picture()
                else:
                    if isinstance(picture, Qt.QPixmap):
                        pixmap = picture
                    elif isinstance(picture, basestring) or isinstance(picture, Qt.QString):
                        picture = str(picture)
                        pixmap = Qt.QPixmap(os.path.realpath(picture))
                    SelectionMark = Qt.QGraphicsPixmapItem()
                    SelectionMark.setPixmap(pixmap.scaled(w, h))
                    SelectionMark.hide()
            except:
                self.debug('In setSelectionMark(%s): %s' %
                           (picture, traceback.format_exc()))
                picture = None
        return SelectionMark

    def drawSelectionMark(self, x, y, w, h, oversize=1):
        """
        If h or w are None the mark is drawn at x,y
        If h or w has a value the mark is drawn in the center of the region ((x,y)(x+w,y+h))
        """
        #self.debug('%s has parent %s' % (item_name,getattr(item.parentItem(),'_name','ung')  if item.parentItem() else 'None'))
        #self.debug('drawSelectionMark(): center and width,height are: (%d,%d),(%d,%d)' % (x,y,w,h))

        mark = self.getSelectionMark()
        self._selection.append(mark)
        srect = self.itemsBoundingRect()
        MAX_CIRCLE_SIZE = srect.width(), srect.height()  # 500,500 #20,20
        LIMITS = (0, 0, srect.width(), srect.height())

        def bound(coords, bounds=LIMITS):
            """ x,y,w,h """
            x, y, w, h = coords
            if x < bounds[0]:
                w, x = w - (bounds[0] - x), bounds[0]
            if y < bounds[1]:
                h, y = h - (bounds[1] - y), bounds[1]
            if x + w > bounds[2]:
                w, x = (bounds[2] - x), x
            if y + h > bounds[3]:
                h, y = (bounds[3] - y), y
            return x, y, w, h

        if isinstance(mark, Qt.QGraphicsEllipseItem):
            if None not in [w, h]:
                if w > MAX_CIRCLE_SIZE[0] or h > MAX_CIRCLE_SIZE[1]:
                    # Applying correction if the file is too big, half max
                    # circle size around the center
                    x, y = (x + w / 2.) - .5 * \
                        MAX_CIRCLE_SIZE[0], (y + h / 2.) - .5 * \
                        MAX_CIRCLE_SIZE[1],
                    w, h = [.5 * t for t in MAX_CIRCLE_SIZE]
                else:
                    x, y = x - .5 * w, y - .5 * h
            else:
                w, h = [.5 * t for t in MAX_CIRCLE_SIZE]
                #self.debug('drawSelectionMark(): center and width,height are: (%d,%d),(%d,%d)' % (x,y,w,h))
            mark.setRect(*bound((x, y, w * 2, h * 2)))
            # mark.setRect(x,y,w*2,h*2)
        elif isinstance(mark, Qt.QGraphicsPixmapItem):
            rect = mark.boundingRect()
            if None not in [w, h]:
                x, y = x + .5 * w, y + .5 * h
            mark.setOffset(x - .5 * rect.width(), y - .5 * rect.height())
        elif isinstance(mark, Qt.QGraphicsItem):
            mark.setRect(x, y, w, h)

        mark.hide()  # It's better to add it hidden to avoid resizings
        self.addItem(mark)
        mark.setZValue(9999)  # Put on Top
        mark.show()
        self.updateSceneViews()
        return

    def getShellCommand(self, obj, wait=False):
        shellCom = obj.getExtensions().get('shellCommand').replace(
            '$NAME', obj._name).replace('$MODEL', obj._name)
        if not wait and not shellCom.endswith('&'):
            shellCom += ' &'
        if obj.noPrompt:
            subprocess.call(shellCom, shell=True)
        else:
            yes = Qt.QMessageBox.Ok
            no = Qt.QMessageBox.Cancel
            result = Qt.QMessageBox.question(self.parent(
            ), "Shell command", "Would you like to call shell command '" + shellCom + "' ?", yes, no)
            if result == yes:
                subprocess.call(shellCom, shell=True)
        return

    def getClass(self, clName):
        if not clName or clName == 'noPanel':
            return None
        elif clName in ('atkpanel.MainPanel', 'atkpanel'):
            clName = 'TaurusDevicePanel'
        # TODO: allow passing class names including module, e.g.: 'foo.Bar'
        if clName in globals():
            return globals()[clName]
        elif clName in locals():
            return locals()[clName]
        elif clName in dir(Qt):
            return getattr(Qt, clName)
        else:
            wf = TaurusWidgetFactory()
            try:
                return wf.getTaurusWidgetClass(clName)
            except:
                return None

    @staticmethod
    def getTaurusParentItem(item, top=True):
        """ Searches within a group hierarchy and returns a parent Taurus component or None if no parent TaurusBaseComponent
            is found."""
        if item is None:
            return None
        first, p = None, item.parentItem()
        while p:
            if isinstance(p, TaurusGraphicsItem):
                if first is None:
                    first = p
                    if not top:
                        break
                elif str(p.getModel()) != str(first.getModel()):
                    break
                else:
                    first = p
            p = p.parentItem()
        return first

    def getAllChildren(self, item, klass=None):
        """ Returns all children elements, filtering by klass if wanted """
        result = []
        try:
            children = item.childItems()
            result.extend(
                c for c in children if not klass or isinstance(c, klass))
            result.extend(c.childItems()
                          for c in children if not klass or isinstance(c, klass))
        except:
            pass
        return result

    def start(self):
        if self.updateThread:
            return
        self.updateQueue = Queue.Queue()
        self.updateThread = TaurusGraphicsUpdateThread(self)
        self.updateThread.start()  # Qt.QThread.HighPriority)

    def getQueue(self):
        return self.updateQueue

    def updateSceneItem(self, item):
        self.updateQueue.put(item)

    def updateSceneItems(self, items):
        self.updateQueue.put(items)

    def updateScene(self):
        self.update()

    def updateSceneViews(self):
        for v in self.views():
            v.viewport().update()
            # v.invalidateScene(self.SelectionCircle.boundingRect())
        return


class QGraphicsTextBoxing(Qt.QGraphicsItemGroup):
    """Display a text inside a virtual box. Support horizontal and vertical alignment"""

    _TEXT_RATIO = 0.8

    def __init__(self, parent=None, scene=None):
        Qt.QGraphicsItemGroup.__init__(self, parent, scene)
        self._rect = Qt.QGraphicsRectItem(self, scene)
        self._rect.setBrush(Qt.QBrush(Qt.Qt.NoBrush))
        self._rect.setPen(Qt.QPen(Qt.Qt.NoPen))
        self._text = Qt.QGraphicsTextItem(self, scene)
        self._text.scale(self._TEXT_RATIO, self._TEXT_RATIO)
        self._validBackground = None
        # using that like the previous code create a worst result
        self.__layoutValide = True
        self._alignment = Qt.Qt.AlignCenter | Qt.Qt.AlignVCenter

    def setRect(self, x, y, width, height):
        self._rect.setRect(x, y, width, height)
        self._invalidateLayout()

    def setPlainText(self, text):
        self._text.setPlainText(text)
        self._invalidateLayout()

    def setValidBackground(self, color):
        self._validBackground = color

    def toPlainText(self):
        return self._text.toPlainText()

    def brush(self):
        return self._rect.brush()

    def setBrush(self, brush):
        self._rect.setBrush(brush)

    def pen(self):
        return self._rect.pen()

    def setPen(self, pen):
        self._rect.setPen(pen)

    def setDefaultTextColor(self, color):
        self._text.setDefaultTextColor(color)

    def setHtml(self, html):
        self._text.setHtml(html)
        self._invalidateLayout()

    def setFont(self, font):
        self._text.setFont(font)
        self._invalidateLayout()

    def setAlignment(self, alignment):
        self._alignment = alignment
        self._invalidateLayout()

    def _invalidateLayout(self):
        """Invalidate the current location of the text"""
        if not self.__layoutValide:
            return
        self.__layoutValide = False
        self.update()

    def _validateLayout(self):
        """Compute the text location"""
        if self.__layoutValide:
            return

        rect = self._rect.rect()
        width, height = rect.width(), rect.height()
        textRect = self._text.boundingRect()

        # horizontal layout
        x = rect.x()
        alignment = int(self._alignment)
        if (alignment & int(Qt.Qt.AlignLeft)) != 0:
            x += 0
        elif (alignment & int(Qt.Qt.AlignHCenter)) != 0:
            x += width * 0.5 - textRect.width() * 0.5 * self._TEXT_RATIO
        elif (alignment & int(Qt.Qt.AlignRight)) != 0:
            x += width - textRect.width() * self._TEXT_RATIO

        # vertical layout
        y = rect.y()
        if (alignment & int(Qt.Qt.AlignTop)) != 0:
            y += 0
        elif (alignment & int(Qt.Qt.AlignVCenter)) != 0:
            y += height * 0.5 - textRect.height() * 0.5 * self._TEXT_RATIO
        elif (alignment & int(Qt.Qt.AlignBottom)) != 0:
            y += height - textRect.height() * self._TEXT_RATIO

        self._text.setPos(x, y)
        self.__layoutValide = True

    def paint(self, painter, option, widget):
        self._validateLayout()
        Qt.QGraphicsItemGroup.paint(self, painter, option, widget)


class QSpline(Qt.QGraphicsPathItem):

    def __init__(self, parent=None, closed=False, control_points=None):
        super(QSpline, self).__init__(parent)
        self.__closed = closed
        if control_points is None:
            control_points = []
        self.setControlPoints(control_points)

    def setControlPoints(self, control_points):
        self.__control_points = control_points
        self.updateSplinePath()

    def setClose(self, isClosed):
        if self.__closed == isClosed:
            return
        self.__closed = isClosed
        self.updateSplinePath()

    def updateSplinePath(self):
        path = Qt.QPainterPath()
        cp = self.__control_points
        nb_points = len(cp)
        if nb_points <= 1:
            pass
        elif nb_points == 2:
            path.moveTo(cp[0])
            path.lineTo(cp[1])
        else:
            path.moveTo(cp[0])
            for i in xrange(1, nb_points - 1, 3):
                p1 = cp[i + 0]
                p2 = cp[i + 1]
                end = cp[i + 2]
                path.cubicTo(p1, p2, end)
            if self.__closed:
                path.lineTo(cp[0])

        self.setPath(path)


class TaurusGraphicsItem(TaurusBaseComponent):
    """Base class for all Taurus Graphics Items"""

    def __init__(self, name=None, parent=None):
        self.call__init__(TaurusBaseComponent, name,
                          parent)  # <- log created here
        #self.debug('TaurusGraphicsItem(%s,%s)' % (name,parent))
        self.ignoreRepaint = False
        self.setName(name)
        self._currFgBrush = None
        self._currBgBrush = None
        self._currText = None
        self._currHtmlText = None
        self._map = None
        self._default = None
        self._visible = None
        # self.getExtensions() <= It must be called AFTER set_common_params()
        # in getGraphicsItem()
        self._contextMenu = []

    def setName(self, name):
        name = str(name or self.__class__.__name__)
        # srubio@cells.es: modified to store ._name since initialization (even
        # if a model is not set)
        self._name = name

    def getName(self):
        return self._name

    def setContextMenu(self, menu):
        '''Context Menu must be a list of tuples (ActionName,ActionMethod), empty tuples insert separators between options.'''
        self._contextMenu = menu

    def contextMenu(self):
        return self._contextMenu

    def getExtensions(self):
        """
        Any in
        ExtensionsList,noPrompt,standAlone,noTooltip,noSelect,ignoreRepaint,shellCommand,className,classParams
        """
        self._extensions = getattr(self, '_extensions', {})
        if 'ExtensionsList' in self._extensions:
            self._extensions.update((k.strip(), True) for k in self._extensions[
                                    'ExtensionsList'].split(','))
            self._extensions.pop('ExtensionsList')
        for k in ('noPrompt', 'standAlone', 'noTooltip', 'ignoreRepaint', 'noSelect'):
            if self._extensions.get(k, None) == '':
                self._extensions[k] = True
        self.noPrompt = self._extensions.get('noPrompt', False)
        self.standAlone = self._extensions.get('standAlone', False)
        self.noTooltip = self._extensions.get('noTooltip', False)
        self.ignoreRepaint = self._extensions.get(
            'ignoreRepaint', self.ignoreRepaint)
        self.setName(self._extensions.get('name', self._name))
        tooltip = '' if (self.noTooltip or self._name ==
                         self.__class__.__name__ or self._name is None) else str(self._name)
        #self.debug('setting %s.tooltip = %s'%(self._name,tooltip))
        self.setToolTip(tooltip)
        #self.debug('%s.getExtensions(): %s'%(self._name,self._extensions))
        return self._extensions

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory methods to be implemented in any subclass of TaurusComponent
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setModel(self, model):
        #self.debug('In %s.setModel(%s)'%(type(self).__name__,model))
        self.setName(model)

        if issubclass(Manager().findObjectClass(self._name), TaurusDevice):
            model = self._name + '/state'
        TaurusBaseComponent.setModel(self, model)

    def getParentTaurusComponent(self):
        """ Returns a parent Taurus component or None if no parent TaurusBaseComponent
            is found."""
        p = self.parentItem()
        while p and not isinstance(p, TaurusGraphicsItem):
            p = self.parentItem()
        return p

    # def fireEvent(self, type):
    def fireEvent(self, evt_src=None, evt_type=None, evt_value=None):
        """fires a value changed event to all listeners"""
        self.updateStyle()

    def updateStyle(self):
        """ Method called when the component detects an event that triggers a change
            in the style."""
        if self.scene():
            self.scene().updateSceneItem(self)

    # getDisplayValue is already (and better) implemented in TaurusBaseComponent
    # def getDisplayValue(self):
        #attrvalue = self.getModelValueObj()
        # if not attrvalue:
            # return self.getNoneValue()
        # return str(attrvalue.value)

    def isReadOnly(self):
        return True

    def __str__(self):
        return self.log_name + "(" + self.modelName + ")"

    def getModelClass(self):
        return TaurusAttribute


class TaurusGraphicsAttributeItem(TaurusGraphicsItem):
    """
    This class show value->text conversion in label widgets.
    Quality is shown in background
    """

    def __init__(self, name=None, parent=None):
        name = name or self.__class__.__name__
        self._unitVisible = True
        self._currValue = None
        self._userFormat = None
        self._unitVisible = True
        self.call__init__(TaurusGraphicsItem, name, parent)

    def getUnit(self):
        unit = ''
        modelObj = self.getModelObj()
        if not modelObj is None:
            unit = modelObj.getUnit()
            if not unit or unit == 'No unit':
                unit = ''
        return unit

    def updateStyle(self):
        v = self.getModelValueObj()
        if self.getShowQuality():
            try:
                quality = None
                if v:
                    quality = v.quality
                if quality == AttrQuality.ATTR_VALID and self._validBackground:
                    background = self._validBackground
                else:
                    background, _ = QT_ATTRIBUTE_QUALITY_PALETTE.qcolor(
                        quality)
                self.setBrush(Qt.QBrush(background))
            except:
                self.warning('In TaurusGraphicsAttributeItem(%s).updateStyle(%s): colors failed!' % (
                    self._name, self._currText))
                self.warning(traceback.format_exc())

        if v and self._userFormat:
            text = self._userFormat % (v.value)
        else:
            text = self._currText = self.getDisplayValue()
        if self._unitVisible:
            text = text + ' ' + self.getUnit()
        self._currText = text.decode('unicode-escape')
        self._currHtmlText = None

        TaurusGraphicsItem.updateStyle(self)

    def setUserFormat(self, format):
        self._userFormat = format

    def setUnitVisible(self, yesno):
        self._unitVisible = yesno


class TaurusGraphicsStateItem(TaurusGraphicsItem):
    """
    In State Item the displayValue should not override the label
    This item will modify only foreground/background colors
    """

    def __init__(self, name=None, parent=None):
        name = name or self.__class__.__name__
        self.call__init__(TaurusGraphicsItem, name, parent)

    def updateStyle(self):
        v = self.getModelValueObj()

        self._currBrush = Qt.QBrush(Qt.Qt.NoBrush)
        if v:  # or self.getShowState():
            try:
                bg_brush, fg_brush = None, None
                if self.getModelObj().getType() == DataType.DevState:
                    bg_brush, fg_brush = QT_DEVICE_STATE_PALETTE.qbrush(
                        v.value)
                elif self.getModelObj().getType() == DataType.Boolean:
                    bg_brush, fg_brush = QT_DEVICE_STATE_PALETTE.qbrush(
                        (DevState.FAULT, DevState.ON)[v.value])
                elif self.getShowQuality():
                    bg_brush, fg_brush = QT_ATTRIBUTE_QUALITY_PALETTE.qbrush(
                        v.quality)
                if None not in (bg_brush, fg_brush):
                    self._currBgBrush = bg_brush
                    self._currFgBrush = fg_brush
                    # If there's no filling, applying background brush to
                    # foreground
                    if Qt.Qt.NoBrush != getattr(self, '_fillStyle', Qt.Qt.NoBrush):
                        #self.debug('In TaurusGraphicsStateItem(%s).updateStyle(%s): switching background to foreground'%(self._name,self._currText))
                        self._currFgBrush = bg_brush
                    if self._currText:
                        self._currHtmlText = '<p style="color:%s">%s</p>' % (
                            self._currBgBrush.color().name(), self._currText)
            except:
                self.warning('In TaurusGraphicsStateItem(%s).updateStyle(%s): colors failed!' % (
                    self._name, self._currText))
                self.warning(traceback.format_exc())

        # Parsing _map to manage visibility (a list of values for which the
        # item is visible or not)
        if (v and not self._map is None and
                self._currText in DevState.__members__):
            #self.debug('In TaurusGraphicsStateItem.updateStyle(): mapping %s'%self._currText)
            if DevState[self._currText] == self._map[1]:
                self.setVisible(self._map[2])
                self._visible = self._map[2]
            else:
                self.setVisible(self._default)
                self._visible = self._default

        TaurusGraphicsItem.updateStyle(self)


class TaurusEllipseStateItem(Qt.QGraphicsEllipseItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        Qt.QGraphicsEllipseItem.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget=None):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        Qt.QGraphicsEllipseItem.paint(self, painter, option, widget)


class TaurusRectStateItem(Qt.QGraphicsRectItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        Qt.QGraphicsRectItem.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        Qt.QGraphicsRectItem.paint(self, painter, option, widget)


class TaurusSplineStateItem(QSpline, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        QSpline.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        QSpline.paint(self, painter, option, widget)


class TaurusRoundRectItem(Qt.QGraphicsPathItem):

    def __init__(self, name=None, parent=None, scene=None):
        Qt.QGraphicsPathItem.__init__(self, parent, scene)
        self.__rect = None
        self.setCornerWidth(0, 0)

    def __updatePath(self):
        if self.__rect is None:
            return
        if self.__corner is None:
            return

        path = Qt.QPainterPath()
        cornerWidth, nbPoints = self.__corner
        if cornerWidth == 0 or nbPoints == 0:
            path.addRect(self.__rect)
        elif cornerWidth * 2 > self.__rect.width():
            path.addRect(self.__rect)
        elif cornerWidth * 2 > self.__rect.height():
            path.addRect(self.__rect)
        else:
            path.addRoundedRect(self.__rect, cornerWidth, cornerWidth)
        self.setPath(path)

    def setRect(self, x, y, width, height):
        self.__rect = Qt.QRectF(x, y, width, height)
        self.__updatePath()

    def setCornerWidth(self, width, nbPoints):
        self.__corner = width, nbPoints
        self.__updatePath()


class TaurusRoundRectStateItem(TaurusRoundRectItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        TaurusRoundRectItem.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        TaurusRoundRectItem.paint(self, painter, option, widget)


class TaurusGroupItem(Qt.QGraphicsItemGroup):

    def __init__(self, name=None, parent=None, scene=None):
        Qt.QGraphicsItemGroup.__init__(self, parent, scene)


class TaurusGroupStateItem(TaurusGroupItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        TaurusGroupItem.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        TaurusGroupItem.paint(self, painter, option, widget)


class TaurusPolygonStateItem(Qt.QGraphicsPolygonItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        #Qt.QGraphicsRectItem.__init__(self, parent, scene)
        Qt.QGraphicsPolygonItem.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        Qt.QGraphicsPolygonItem.paint(self, painter, option, widget)


class TaurusLineStateItem(Qt.QGraphicsLineItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        Qt.QGraphicsLineItem.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        Qt.QGraphicsLineItem.paint(self, painter, option, widget)


class TaurusTextStateItem(QGraphicsTextBoxing, TaurusGraphicsStateItem):
    """
    A QGraphicsItem that represents a text related to a device state or attribute quality
    """

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        QGraphicsTextBoxing.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currHtmlText:
            self.setHtml(self._currHtmlText)
        else:
            self.setPlainText(self._currText or '')
        QGraphicsTextBoxing.paint(self, painter, option, widget)


class TaurusTextAttributeItem(QGraphicsTextBoxing, TaurusGraphicsAttributeItem):
    """
    A QGraphicsItem that represents a text related to an attribute value
    """

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        QGraphicsTextBoxing.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsAttributeItem, name, parent)

    def paint(self, painter, option, widget):
        # self.debug('TaurusTextAttributeItem(%s,%s,%s)'%(self.getName(),self._currText,self._currHtmlText))
        if self._currHtmlText:
            self.setHtml(self._currHtmlText)
        else:
            self.setPlainText(self._currText or '')
        QGraphicsTextBoxing.paint(self, painter, option, widget)


TYPE_TO_GRAPHICS = {
    None: {"Rectangle": Qt.QGraphicsRectItem,
           "RoundRectangle": TaurusRoundRectItem,
           "Ellipse": Qt.QGraphicsEllipseItem,
           "Polyline": Qt.QGraphicsPolygonItem,
           "Label": QGraphicsTextBoxing,
           "Line": Qt.QGraphicsLineItem,
           "Group": TaurusGroupItem,
           "SwingObject": TaurusTextAttributeItem,
           "Image": Qt.QGraphicsPixmapItem,
           "Spline": QSpline, },

    TaurusDevice: {"Rectangle": TaurusRectStateItem,
                   "RoundRectangle": TaurusRoundRectStateItem,
                   "Ellipse": TaurusEllipseStateItem,
                   "Polyline": TaurusPolygonStateItem,
                   "Label": TaurusTextStateItem,
                   "Line": Qt.QGraphicsLineItem,  # TaurusLineStateItem,
                   "Group": TaurusGroupStateItem,
                   "SwingObject": TaurusTextAttributeItem,
                   "Image": Qt.QGraphicsPixmapItem,
                   "Spline": TaurusSplineStateItem, },

    TaurusAttribute: {"Rectangle": TaurusRectStateItem,
                      "RoundRectangle": TaurusRoundRectStateItem,
                      "Ellipse": TaurusEllipseStateItem,
                      "Polyline": TaurusPolygonStateItem,
                      "Label": TaurusTextAttributeItem,
                      "Line": Qt.QGraphicsLineItem,  # TaurusLineStateItem,
                      "Group": TaurusGroupStateItem,
                      "SwingObject": TaurusTextAttributeItem,
                      "Image": Qt.QGraphicsPixmapItem,
                      "Spline": TaurusSplineStateItem, },
}


class TaurusBaseGraphicsFactory:

    def __init__(self):
        pass

    def getSceneObj(self):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getSceneObj()")

    def getObj(self, name, params):
        raise RuntimeError("Invalid call to AbstractGraphicsFactory::getObj()")

    def getRectangleObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getRectangleObj()")

    def getRoundRectangleObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getRoundRectangleObj()")

    def getLineObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getLineObj()")

    def getEllipseObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getEllipseObj()")

    def getPolylineObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getPolylineObj()")

    def getLabelObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getLabelObj()")

    def getGroupObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getGroupObj()")

    def getSwingObjectObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getSwingObjectObj()")

    def getImageObj(self, parms):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getImageObj()")

    def getSplineObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getSplineObj()")

    def getGraphicsClassItem(self, cls, type_):
        ncls = cls
        try:
            if issubclass(cls, TaurusDevice):
                ncls = TaurusDevice
            elif issubclass(cls, TaurusAttribute):
                ncls = TaurusAttribute
        except:
            pass
        ncls = TYPE_TO_GRAPHICS.get(
            ncls, TYPE_TO_GRAPHICS.get(None)).get(type_)
        return ncls

    def getGraphicsItem(self, type_, params):
        name = params.get(self.getNameParam())
        # applying alias
        for k, v in getattr(self, 'alias', {}).items():
            if k in name:
                name = str(name).replace(k, v)
                params[self.getNameParam()] = name
        cls = None
        if '/' in name:
            # replacing Taco identifiers in %s'%name
            if name.lower().startswith('tango:') and (name.count('/') == 2 or not 'tango:/' in name.lower()):
                nname = name.split(':', 1)[-1]
                params[self.getNameParam()] = name = nname
            if name.lower().endswith('/state'):
                name = name.rsplit('/', 1)[0]
            cls = Manager().findObjectClass(name)
        else:
            if name:
                self.debug('%s does not match a tango name' % name)
        klass = self.getGraphicsClassItem(cls, type_)
        item = klass()
        # It's here were Attributes are subscribed
        self.set_common_params(item, params)
        if hasattr(item, 'getExtensions'):
            item.getExtensions()  # <= must be called here to take extensions from params
        return item

    def getNameParam(self):
        """Returns the name of the parameter which contains the name identifier.
           Default implementation returns 'name'.
           Overwrite has necessary."""
        return 'name'

    def set_common_params(self, item, params):
        """Sets the common parameters. Default implementation does nothing.
           Overwrite has necessary."""
        pass



# -*- coding: utf-8 -*-
"""
    pint.systems
    ~~~~~~~~~~~~

    Functions and classes related to system definitions and conversions.

    :copyright: 2013 by Pint Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import division, unicode_literals, print_function, absolute_import

import re

from .unit import Definition, UnitDefinition, DefinitionSyntaxError
from .util import to_units_container, UnitsContainer


class Group(object):
    """A group is a list of units.

    Units can be added directly or by including other groups.

    Members are computed dynamically, that is if a unit is added to a group X
    all groups that include X are affected.
    """

    #: Regex to match the header parts of a context.
    _header_re = re.compile('@group\s+(?P<name>\w+)\s*(using\s(?P<used_groups>.*))*')

    def __init__(self, name, groups_systems):
        """
        :param name: Name of the group
        :type name: str
        :param groups_systems: dictionary containing groups and system.
                               The newly created group will be added after creation.
        :type groups_systems: dict[str, Group | System]
        """

        if name in groups_systems:
            t = 'group' if isinstance(groups_systems['name'], Group) else 'system'
            raise ValueError('The system name already in use by a %s' % t)

        # The name of the group.
        #: type: str
        self.name = name

        #: Names of the units in this group.
        #: :type: set[str]
        self._unit_names = set()

        #: Names of the groups in this group.
        #: :type: set[str]
        self._used_groups = set()

        #: Names of the groups in which this group is contained.
        #: :type: set[str]
        self._used_by = set()

        #: Maps group name to Group.
        #: :type: dict[str, Group]
        self._groups_systems = groups_systems

        self._groups_systems[self.name] = self

        if name != 'root':
            # All groups are added to root group
            groups_systems['root'].add_groups(name)

        #: A cache of the included units.
        #: None indicates that the cache has been invalidated.
        #: :type: frozenset[str] | None
        self._computed_members = None

    @property
    def members(self):
        """Names of the units that are members of the group.

        Calculated to include to all units in all included _used_groups.

        :rtype: frozenset[str]
        """
        if self._computed_members is None:
            self._computed_members = set(self._unit_names)

            for _, group in self.iter_used_groups():
                self._computed_members |= group.members

            self._computed_members = frozenset(self._computed_members)

        return self._computed_members

    def invalidate_members(self):
        """Invalidate computed members in this Group and all parent nodes.
        """
        self._computed_members = None
        d = self._groups_systems
        for name in self._used_by:
            d[name].invalidate_members()

    def iter_used_groups(self):
        pending = set(self._used_groups)
        d = self._groups_systems
        while pending:
            name = pending.pop()
            group = d[name]
            pending |= group._used_groups
            yield name, d[name]

    def is_used_group(self, group_name):
        for name, _ in self.iter_used_groups():
            if name == group_name:
                return True
        return False

    def add_units(self, *unit_names):
        """Add units to group.

        :type unit_names: str
        """
        for unit_name in unit_names:
            self._unit_names.add(unit_name)

        self.invalidate_members()

    def remove_units(self, *unit_names):
        """Remove units from group.

        :type unit_names: str
        """
        for unit_name in unit_names:
            self._unit_names.remove(unit_name)

        self.invalidate_members()

    def add_groups(self, *group_names):
        """Add groups to group.

        :type group_names: str
        """
        d = self._groups_systems
        for group_name in group_names:

            grp = d[group_name]

            if grp.is_used_group(self.name):
                raise ValueError('Cyclic relationship found between %s and %s' % (self.name, group_name))

            self._used_groups.add(group_name)
            grp._used_by.add(self.name)

        self.invalidate_members()

    def remove_groups(self, *group_names):
        """Remove groups from group.

        :type group_names: str
        """
        d = self._groups_systems
        for group_name in group_names:
            grp = d[group_name]

            self._used_groups.remove(group_name)
            grp._used_by.remove(self.name)


        self.invalidate_members()

    @classmethod
    def from_lines(cls, lines, define_func, group_dict):
        """Return a Group object parsing an iterable of lines.

        :param lines: iterable
        :type lines: list[str]
        :param define_func: Function to define a unit in the registry.
        :type define_func: str -> None
        :param group_dict: Maps group name to Group.
        :type group_dict: dict[str, Group]
        """
        header, lines = lines[0], lines[1:]

        r = cls._header_re.search(header)
        name = r.groupdict()['name'].strip()
        groups = r.groupdict()['used_groups']
        if groups:
            group_names = tuple(a.strip() for a in groups.split(','))
        else:
            group_names = ()

        unit_names = []
        for line in lines:
            if '=' in line:
                # Is a definition
                definition = Definition.from_string(line)
                if not isinstance(definition, UnitDefinition):
                    raise DefinitionSyntaxError('Only UnitDefinition are valid inside _used_groups, '
                                                'not %s' % type(definition))
                define_func(definition)
                unit_names.append(definition.name)
            else:
                unit_names.append(line.strip())

        grp = cls(name, group_dict)

        grp.add_units(*unit_names)

        if group_names:
            grp.add_groups(*group_names)

        return grp


class System(object):
    """A system is a Group plus a set of base units.

    @system <name> [using <group 1>, ..., <group N>]
        <rule 1>
        ...
        <rule N>
    @end

    """

    #: Regex to match the header parts of a context.
    _header_re = re.compile('@system\s+(?P<name>\w+)\s*(using\s(?P<used_groups>.*))*')

    def __init__(self, name, groups_systems):
        """
        :param name: Name of the group
        :type name: str
        :param groups_systems: dictionary containing groups and system.
                               The newly created group will be added after creation.
        :type groups_systems: dict[str, Group | System]
        """

        if name in groups_systems:
            t = 'group' if isinstance(groups_systems[name], Group) else 'system'
            raise ValueError('The system name (%s) already in use by a %s' % (name, t))

        #: Name of the system
        #: :type: str
        self.name = name

        #: Maps root unit names to a dict indicating the new unit and its exponent.
        #: :type: dict[str, dict[str, number]]]
        self.base_units = {}

        #: Derived unit names.
        #: :type: set(str)
        self.derived_units = set()

        #: Names of the _used_groups in used by this system.
        #: :type: set(str)
        self._used_groups = set()

        #: :type: frozenset | None
        self._computed_members = None

        #: Maps group name to Group.
        self._group_systems_dict = groups_systems

        self._group_systems_dict[self.name] = self

    @property
    def members(self):
        if self._computed_members is None:
            self._computed_members = set()

            for group_name in self._used_groups:
                self._computed_members |= self._group_systems_dict[group_name].members

            self._computed_members = frozenset(self._computed_members)

        return self._computed_members

    def invalidate_members(self):
        """Invalidate computed members in this Group and all parent nodes.
        """
        self._computed_members = None

    def add_groups(self, *group_names):
        """Add groups to group.

        :type group_names: str
        """
        self._used_groups |= set(group_names)

        self.invalidate_members()

    def remove_groups(self, *group_names):
        """Remove groups from group.

        :type group_names: str
        """
        self._used_groups -= set(group_names)

        self.invalidate_members()

    @classmethod
    def from_lines(cls, lines, get_root_func, group_dict):
        header, lines = lines[0], lines[1:]

        r = cls._header_re.search(header)
        name = r.groupdict()['name'].strip()
        groups = r.groupdict()['used_groups']

        # If the systems has no group, it automatically uses the root group.
        if groups:
            group_names = tuple(a.strip() for a in groups.split(','))
        else:
            group_names = ('root', )

        base_unit_names = {}
        derived_unit_names = []
        for line in lines:
            line = line.strip()

            # We would identify a
            #  - old_unit: a root unit part which is going to be removed from the system.
            #  - new_unit: a non root unit which is going to replace the old_unit.

            if ':' in line:
                # The syntax is new_unit:old_unit

                new_unit, old_unit = line.split(':')
                new_unit, old_unit = new_unit.strip(), old_unit.strip()

                # The old unit MUST be a root unit, if not raise an error.
                if old_unit != str(get_root_func(old_unit)[1]):
                    raise ValueError('In `%s`, the unit at the right of the `:` must be a root unit.' % line)

                # Here we find new_unit expanded in terms of root_units
                new_unit_expanded = to_units_container(get_root_func(new_unit)[1])

                # We require that the old unit is present in the new_unit expanded
                if old_unit not in new_unit_expanded:
                    raise ValueError('Old unit must be a component of new unit')

                # Here we invert the equation, in other words
                # we write old units in terms new unit and expansion
                new_unit_dict = dict((new_unit, -1./value)
                                     for new_unit, value in new_unit_expanded.items()
                                     if new_unit != old_unit)
                new_unit_dict[new_unit] = 1 / new_unit_expanded[old_unit]

                base_unit_names[old_unit] = new_unit_dict

            else:
                # The syntax is new_unit
                # old_unit is inferred as the root unit with the same dimensionality.

                new_unit = line
                old_unit_dict = to_units_container(get_root_func(line)[1])

                if len(old_unit_dict) != 1:
                    raise ValueError('The new base must be a root dimension if not discarded unit is specified.')

                old_unit, value = dict(old_unit_dict).popitem()

                base_unit_names[old_unit] = {new_unit: 1./value}

        system = cls(name, group_dict)

        system.add_groups(*group_names)

        system.base_units.update(**base_unit_names)
        system.derived_units |= set(derived_unit_names)

        return system


class GSManager(object):

    def __init__(self):
        #: :type: dict[str, Group | System]
        self._groups_systems = dict()
        self._root_group = Group('root', self._groups_systems)

    def add_system_from_lines(self, lines, get_root_func):
        System.from_lines(lines, get_root_func, self._groups_systems)

    def add_group_from_lines(self, lines, define_func):
        Group.from_lines(lines, define_func, self._groups_systems)

    def get_group(self, name, create_if_needed=True):
        """Return a Group.

        :param name: Name of the group.
        :param create_if_needed: Create a group if not Found. If False, raise an Exception.
        :return: Group
        """
        try:
            return self._groups_systems[name]
        except KeyError:
            if create_if_needed:
                if name == 'root':
                    raise ValueError('The name root is reserved.')
                return Group(name, self._groups_systems)
            else:
                raise KeyError('No group %s found.' % name)

    def get_system(self, name, create_if_needed=True):
        """Return a Group.

        :param name: Name of the system
        :param create_if_needed: Create a group if not Found. If False, raise an Exception.
        :return: System
        """

        try:
            return self._groups_systems[name]
        except KeyError:
            if create_if_needed:
                return System(name, self._groups_systems)
            else:
                raise KeyError('No system found named %s.' % name)

    def __getitem__(self, item):
        if item in self._groups_systems:
            return self._groups_systems[item]

        raise KeyError('No group or system found named %s.' % item)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

# ----------------------------------------------------------------------------
# This class also borrows much code from the build_mock_qt.py script from
# qarbon (http://qarbon.rtfd.org/)
#
# Copyright (c) 2013 European Synchrotron Radiation Facility, Grenoble, France
#
# ----------------------------------------------------------------------------

''' Creates a tree of dirs and restructured text stub files for documenting
the API of a python module with sphinx'''

from __future__ import with_statement

import os
import sys
import glob
import re
import inspect
import shutil

# Define templates

module_init_template = """\
from __future__ import print_function

{imports}

class _MockMeta(type):
    def __getattr__(self, name):
        return _Mock()

class _Mock(object):
    __metaclass__ = _MockMeta
    def __init__(self, *a, **kw):
        object.__init__(self)
        for k,v in kw.iteritems():
            setattr(self, k, v)
    def __getattr__(*a, **kw): return _Mock()
    def __call__(*a, **kw): return _Mock()
    def __getitem__(*a, **kw): return _Mock()
    def __int__(*a, **kw): return 1
    def __contains__(*a, **kw): return False
    def __len__(*a, **kw): return 1
    def __iter__(*a, **kw): return iter([])
    def __exit__(*a, **kw): return False
    def __complex__(*a, **kw): return 1j
    def __float__(*a, **kw): return 1.0
    def __bool__(*a, **kw): return True
    def __nonzero__(*a, **kw): return True
    def __oct__(*a, **kw): return 1
    def __hex__(*a, **kw): return 0x1
    def __long__(*a, **kw): return long(1)
    def __index__(*a, **kw): return 1
"""

import_template = """import {name} as {asname}"""
mock_template = """{name} = _Mock()"""

klass_template = """\
class {klass}({super_klass}):
  pass
{members}"""

function_template = """def {function}(*a,**k): return _Mock()"""
member_template = """  {name} = {value!r}"""
constant_template = """{name} = {value!r}"""
specialfloats_template = """{name} = float('{value!r}')"""


def abspath(*path):
    """A method to determine absolute path for a given relative path to the
    directory where this .py script is located"""
    this_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.abspath(os.path.join(this_dir, *path))


def _import(name):
    __import__(name)
    return sys.modules[name]


def _is_pseudo_enum(obj):
    if not isinstance(obj, int):
        return False
    try:
        int(repr(obj))
        return False
    except:
        return True


def _is_special_float(obj):
    if not isinstance(obj, float):
        return False
    return repr(obj) in ('inf', 'nan', '-inf')


def _discard_element(name, exclude, include):
    if name in include:
        return False
    return name in exclude or name.startswith("__")


def build_class(k_name, k, exclude=(), include=()):
    '''return the source text for a mock class based on a given class'''
    methods = []
    members = []

    for element_name in dir(k):
        if _discard_element(element_name, exclude, include):
            continue
        try:
            element = getattr(k, element_name)
        except AttributeError:
            continue
        if _is_special_float(element):
            members.append(specialfloats_template.format(name=element_name,
                                                         value=element))
        elif isinstance(element, (int, float, bool, str, unicode)):
            try:  # make sure that the repr makes sense
                type(element)(repr(element))
            except:  # skip it (the _Mock.__getattr__ will deal with it)
                continue
            members.append(member_template.format(name=element_name,
                                                  value=element))
    members = "\n".join(members)

    klass_str = klass_template.format(klass=k_name,
                                      super_klass='_Mock',
                                      members=members)
    return klass_str


def build_module(module_name, imports=(), out_prefix='mock',
                 exclude=(), include=()):
    '''generate a mock package for a given module'''
    rel_dir = module_name.split(".")
    abs_dir = abspath(out_prefix, *rel_dir)
    if os.path.isdir(abs_dir):
        return
    os.makedirs(abs_dir)
    module = _import(module_name)
    fake_module_filename = os.path.join(abs_dir, "__init__.py")

    klasses = []
    constants = []
    mocks = []
    imports = set([import_template.format(name=m, asname=m) for m in imports])
    for element_name in sorted(dir(module)):
        if _discard_element(element_name, exclude, include):
            continue
        element = getattr(module, element_name)
        # internal imports (from the same package)
        if (inspect.ismodule(element) and
                element.__name__.split('.')[0] == module_name.split('.')[0]):
            # add the module to the imports set #@TODO: this does not work
            # imports.add(import_template.format(name=element.__name__,
            #                                    asname=element_name))
            # @todo: The above lines were commented because they created
            # problems with circular imports. So for now we just mock them
            mocks.append(mock_template.format(name=element_name))
            # make sure that the module is built
            build_module(element.__name__, imports=(),
                         out_prefix=out_prefix, exclude=exclude,
                         include=include)
        # classes
        elif inspect.isclass(element):
            klasses.append(build_class(element_name, element,
                                       exclude=exclude, include=include))
        # inf, and NaN constants
        elif _is_special_float(element):
            constants.append(specialfloats_template.format(name=element_name,
                                                           value=element))
        # enumerations-like objects
        elif (_is_pseudo_enum(element)):
            constants.append(mock_template.format(name=element_name))
        # constants
        elif isinstance(element, (int, float, bool, str, unicode)):
            try:  # make sure that the repr makes sense
                type(element)(repr(element))
            except:  # cannot write anything better than a mock
                constants.append(mock_template.format(name=element_name))
            constants.append(constant_template.format(name=element_name,
                                                      value=element))
        # final catch-all: it covers modules, functions and other elements
        # that aren't caught by any of the above
        elif (element_name not in imports):
            mocks.append(mock_template.format(name=element_name))

    imports = "\n".join(sorted(imports))

    module_init = module_init_template.format(imports=imports)
    mocks = "\n\n".join(mocks)
    constants = "\n\n".join(constants)
    klasses = "\n\n".join(klasses)
    with open(fake_module_filename, "w") as f:
        f.write(module_init)
        f.write("\n\n")
        f.write(mocks)
        f.write("\n\n")
        f.write(constants)
        f.write("\n\n")
        f.write(klasses)
        f.write("\n\n")


def guess_submodules_from_package(module_name, exclude=(), include=()):
    '''returns a list of submodule names found in a given package name.
    If module_name is not implemented as a package, it returns an empty list'''
    if module_name in exclude:
        return []
    module = _import(module_name)
    try:
        modulefile = inspect.getfile(module)
    except TypeError:
        return []
    if not (modulefile.endswith('__init__.py') or
            modulefile.endswith('__init__.pyc')):
        return []

    pkgdir, _ = os.path.split(modulefile)
    # explore pkgdir to find subdirs with __init__.py files
    g = glob.glob(os.path.join(pkgdir, '*', '__init__.py'))
    names = [re.findall(r".+\/(.*)\/__init__.py", s)[0] for s in g]
    # explore pkgdir to find .py files
    g = glob.glob(os.path.join(pkgdir, '*.py'))
    names += [re.findall(r".+\/(.*).py", s)[0] for s in g]
    # explore pkgdir to find .pyc files
    g = glob.glob(os.path.join(pkgdir, '*.pyc'))
    names += [re.findall(r".+\/(.*).pyc", s)[0] for s in g]
    # explore pkgdir to find .so files
    g = glob.glob(os.path.join(pkgdir, '*.so'))
    names += [re.findall(r".+\/(.*).so", s)[0] for s in g]
    # build list with full module names and filter out non-importable
    # submodules
    full_module_names = []
    for sm_name in names:
        name = '.'.join((module_name, sm_name))
        # skip __main__ and __init__, etc and excluded (unless included)
        if (name not in include and
                (name in exclude or sm_name.startswith('__'))):
            continue
        # check if the module is indeed importable
        try:
            print name
            _import(name)
            full_module_names.append(name)
        except:
            print '!'
            pass
    return full_module_names


def build_full_module(module_name, exclude=(), include=(), out_prefix='mock'):
    '''build a full mocked package (modules and submodules, recursively) for the
    given module'''
    rel_dir = module_name.split(".")
    abs_dir = abspath(out_prefix, *rel_dir)
    if os.path.isdir(abs_dir):
        shutil.rmtree(abs_dir)

    build_module(module_name, imports=(), exclude=exclude, include=include,
                 out_prefix=out_prefix)

    # recursive call for submodules
    for name in guess_submodules_from_package(module_name, exclude=exclude):
        build_full_module(name, exclude=exclude, include=include,
                          out_prefix=out_prefix)


def _zipdir(basedir, archivename):
    '''function to zip directories. Adapted from:
    http://stackoverflow.com/questions/296499
    '''
    from zipfile import ZipFile, ZIP_DEFLATED
    from contextlib import closing
    assert os.path.isdir(basedir)
    with closing(ZipFile(archivename, "w", ZIP_DEFLATED)) as z:
        for root, dirs, files in os.walk(basedir):
            # NOTE: ignore empty directories
            for fn in files:
                absfn = os.path.join(root, fn)
                zfn = absfn[len(basedir) + len(os.sep):]  # XXX: relative path
                z.write(absfn, zfn)


def build_mocks_for_taurus(output='mock.zip'):
    '''builds mocks for the packages required by taurus. The mocks are written
    into the given output directory (or a zip file if output ends with ".zip")
    '''
    import sys
    import tempfile

    if output.endswith('.zip'):
        zfile, outdir = output, tempfile.mkdtemp()
    else:
        zfile, outdir = None, output

    module_names = ['PyTango', 'PyMca5', 'numpy', 'PyQt4', 'sip', 'lxml',
                    'guidata', 'guiqwt', 'spyderlib', 'IPython', 'ply']
    #module_names = ['numpy']

    exclude = ['exec', 'None',
               'spyderlib.scientific_startup',
               'spyderlib.spyder',
               'spyderlib.widgets.externalshell.start_ipython_kernel']
    include = ['__version__']

    for module_name in module_names:
        build_full_module(module_name, exclude=exclude, include=include,
                          out_prefix=outdir)
    if zfile:
        _zipdir(outdir, zfile)  # compress the dir into the zip file
        shutil.rmtree(outdir)  # delete the dir
    print '\nMocks written in %s' % output


if __name__ == "__main__":
    build_mocks_for_taurus()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

""" Every TaurusWidget should have the following Qt Designer extended capabilities:

  - Task menu:
    it means when you right click on the widget in the designer, it will have
    the following additional items:
    - 'Edit model...' - opens a customized dialog for editing the widget model

  - Property Sheet:
    it means that in the Qt Designer property sheet it will have the following
    properties customized:
    - 'model' - will have a '...' button that will open a customized dialog for
      editing the widget model (same has 'Edit model...' task menu item
"""

from taurus.core.util.log import Logger
from taurus.external.qt import Qt
from taurus.external.qt import QtDesigner

from taurus.qt.qtgui.container.qcontainer import QGroupWidget

#import sip

Q_TYPEID = {'QPyDesignerContainerExtension':     'com.trolltech.Qt.Designer.Container',
            'QPyDesignerPropertySheetExtension': 'com.trolltech.Qt.Designer.PropertySheet',
            'QPyDesignerTaskMenuExtension':      'com.trolltech.Qt.Designer.TaskMenu',
            'QPyDesignerMemberSheetExtension':   'com.trolltech.Qt.Designer.MemberSheet'}


class QGroupWidgetContainerExtension(QtDesigner.QPyDesignerContainerExtension):

    def __init__(self, widget, parent=None):
        super(QGroupWidgetContainerExtension, self).__init__(parent)
        self._widget = widget
        self._page_widget = None

    def addWidget(self, widget):
        if self.count() > 0:
            raise Exception("Can only have at maximum one child")
        self._layout().addWidget(widget)
        self._page_widget = widget

    def _content(self):
        return self._widget.content()

    def _layout(self):
        return self._content().layout()

    def count(self):
        return self._layout().count()

    def currentIndex(self):
        if self.count() > 0:
            return 0
        return -1

    def insertWidget(self, index, widget):
        self.addWidget(widget)

    def remove(self, index):
        self._layout().removeWidget(self.widget(index))

    def setCurrentIndex(self, index):
        pass

    def widget(self, index):
        return self._page_widget


class QGroupWidgetExtensionFactory(QtDesigner.QExtensionFactory):

    def __init__(self, parent=None):
        super(QGroupWidgetExtensionFactory, self).__init__(parent)

    def createExtension(self, obj, iid, parent):
        if iid != Q_TYPEID['QPyDesignerContainerExtension']:
            return None
        if isinstance(obj, QGroupWidget):
            return QGroupWidgetContainerExtension(obj, parent)
        return None


def create_plugin():
    from taurusplugin.taurusplugin import TaurusWidgetPlugin

    class QGroupWidgetPlugin(TaurusWidgetPlugin):

        WidgetClass = QGroupWidget

        def initialize(self, formEditor):
            if self.isInitialized():
                return

            manager = formEditor.extensionManager()
            if manager:
                self.factory = QGroupWidgetExtensionFactory(manager)
                manager.registerExtensions(self.factory, Q_TYPEID[
                                           'QPyDesignerContainerExtension'])
            self.initialized = True
    return QGroupWidgetPlugin

QGroupWidgetPlugin = create_plugin()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Taurus provides a framework for testing.
This framework intends to facilitate evaluation, bug finding and integration of
contributed code/patches, as well as to promote test driven development in
Taurus.

The first implementation of this Framework is an outcome of the [Sardana
Enhancement Proposal 5 (SEP5)](http://sourceforge.net/p/sardana/wiki/SEP5/)

Ideally, bug reports should be accompanied by a test revealing the bug,
whenever possible.

The first tests implemented are focused on Unit Tests, but the same framework
should be used for integration and system tests as well.

The taurus.test.testsuite module provides an autodiscovered suite for all
tests implemented in Taurus.

The following are some key points to keep in mind when using this framework:

- The Taurus test framework is based on :mod:`unittest` which should be imported
  from :mod:`taurus.external` in order to be compatible with all versions of
  python supported by Taurus.

- all test-related code is contained in submodules named `test` which appear
  in any module of taurus.

- test-related code falls in one of these three categories:

  - actual test code (classes that derive from unittest.TestCase)

  - utility classes/functions (code to simplify development of test code)

  - resources (accessory files required by some test). They are located in
    subdirectories named `res`

For a more complete description of the conventions on how to write tests with
the Taurus testing framework, please refer to the
[SEP5](http://sourceforge.net/p/sardana/wiki/SEP5/).
"""

from .moduleexplorer import ModuleExplorer
from .resource import getResourcePath
from .base import insertTest
from .fuzzytest import calculateTestFuzziness, loopSubprocess, loopTest



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

'''
Epics module. See __init__.py for more detailed documentation
'''
__all__ = ['EpicsFactory']


import weakref

try:
    import epics
except ImportError:
    from taurus.core.util.log import debug
    debug('cannot import epics module. ' +
          'Taurus will not support the "epics" scheme')
    raise

from taurus.core.taurusexception import TaurusException
from taurus.core.util.singleton import Singleton
from taurus.core.util.log import Logger
from taurus.core.taurusbasetypes import TaurusElementType
from taurus.core.taurusfactory import TaurusFactory

from epicsattribute import EpicsAttribute
from epicsdevice import EpicsDevice
from epicsauthority import EpicsAuthority


class EpicsFactory(Singleton, TaurusFactory, Logger):
    """
    A Singleton class that provides Epics related objects.
    """

    schemes = ("ca", "epics",)
    DEFAULT_DEVICE = 'ca:'
    DEFAULT_AUTHORITY = 'ca://'
    caseSensitive = False
    elementTypesMap = {TaurusElementType.Authority: EpicsAuthority,
                       TaurusElementType.Device: EpicsDevice,
                       TaurusElementType.Attribute: EpicsAttribute
                       }

    def __init__(self):
        """ Initialization. Nothing to be done here for now."""
        pass

    def init(self, *args, **kwargs):
        """Singleton instance initialization."""
        name = self.__class__.__name__
        self.call__init__(Logger, name)
        self.call__init__(TaurusFactory)
        if EpicsAttribute is None:
            self.debug('cannot import epics module.'
                       'Taurus will not support the "epics" scheme')
            raise Exception('"epics" module is not available')
        self.epics_attrs = weakref.WeakValueDictionary()
        self.epics_devs = weakref.WeakValueDictionary()

    def getAuthority(self, name=None):
        """Obtain the Epics (ca) authority object.

        :param name: (str) only a dummy authority ("ca://") is supported

        :return: (EpicsAuthority)
        """
        if name is None:
            name = 'ca://'

        v = self.getAuthorityNameValidator()
        if not v.isValid(name):
            raise TaurusException("Invalid ca authority name %s" % name)

        if not hasattr(self, "_auth"):
            self._auth = EpicsAuthority(self.DEFAULT_AUTHORITY)
        return self._auth

    def getDevice(self, dev_name):
        """Obtain the EpicsDevice object.

        :param dev_name: (str) only one dummy device ("") is supported

        :return: (EpicsDevice)

        .. todo:: epics.Device may be wrapped as taurus device...
        """
        validator = self.getDeviceNameValidator()
        names = validator.getNames(dev_name)
        if names is None:
            raise TaurusException("Invalid epics device name %s" % dev_name)
        fullname = names[0]
        d = self.epics_devs.get(fullname, None)
        if d is None:  # if the full name is not there, create one
            auth = self.getAuthority()
        d = EpicsDevice(fullname, parent=auth)
        self.epics_devs[fullname] = d
        return d

    def getAttribute(self, attr_name):
        """Obtain the object corresponding to the given attribute name. If the
        corresponding attribute already exists, the existing instance is
        returned. Otherwise a new instance is stored and returned. The
        device associated to this attribute will also be created if necessary.

        :param attr_name: (str) the attribute name string. See
                          :mod:`taurus.core.epics` for valid attribute names

        :return: (EpicsAttribute)

        :raises: :TaurusException: if the given name is invalid.
        """
        validator = self.getAttributeNameValidator()
        names = validator.getNames(attr_name)
        if names is None:
            raise TaurusException(
                "Invalid epics attribute name %s" % attr_name)
        fullname = names[0]
        a = self.epics_attrs.get(fullname, None)
        if a is None:  # if the full name is not there, create one
            a = EpicsAttribute(fullname, parent=None)  # note: no parent!
            self.epics_attrs[fullname] = a
        return a

    def getAuthorityNameValidator(self):
        """Return EpicsAuthorityNameValidator"""
        import epicsvalidator
        return epicsvalidator.EpicsAuthorityNameValidator()

    def getDeviceNameValidator(self):
        """Return EpicsDeviceNameValidator"""
        import epicsvalidator
        return epicsvalidator.EpicsDeviceNameValidator()

    def getAttributeNameValidator(self):
        """Return EpicsAttributeNameValidator"""
        import epicsvalidator
        return epicsvalidator.EpicsAttributeNameValidator()

if __name__ == "__main__":
    pass
